# Requirements: NOTE: Change the torch version according to your CUDA version
# python >=3.8
# Step 1: Use pip to install torch==2.2.2 torchvision==0.17.2 torchaudio==2.2.2:
# go to https://docs.pytorch.org/get-started/previous-versions/, search for version 2.2.2 and find your system and GPU version
# For example, if system is Linux and GPU version is CUDA 12.1, run this:
# pip install torch==2.2.2 torchvision==0.17.2 torchaudio==2.2.2 --index-url https://download.pytorch.org/whl/cu121
# Step 2:
# pip install torch_geometric==2.3.0 transformers==4.38
#!/usr/bin/env python

# Suppress known benign warnings from dependencies
import warnings
warnings.filterwarnings('ignore', category=FutureWarning, module='huggingface_hub.file_download')
warnings.filterwarnings('ignore', message='.*Some weights of EsmModel were not initialized.*')
warnings.filterwarnings('ignore', message='.*You should probably TRAIN this model.*')

# Also suppress transformers warnings at import time
import os
os.environ['TRANSFORMERS_NO_ADVISORY_WARNINGS'] = '1'

# Suppress all transformers warnings
import logging
logging.getLogger("transformers").setLevel(logging.ERROR)

import copy
import math
import torch
import pickle
import os
import re
from itertools import chain

from torch_geometric.nn.models import AttentiveFP
from torch.optim import Adam
from torch.nn import MSELoss
from torch.nn import L1Loss
import torch.nn.functional as F

from torch_geometric.data import Data
from torch_geometric.loader import DataLoader
import numpy as np

np.random.seed(0)
torch.manual_seed(0)

# Module-level cache for ESM2 model and tokenizer to avoid reloading
_ESM2_MODEL_CACHE = None
_ESM2_TOKENIZER_CACHE = None

# For parsing PDB files
aminoacid_abbr: dict[str, str] = {
    'GLY': 'G', 'ALA': 'A', 'VAL': 'V', 'LEU': 'L', 'ILE': 'I',
    'PHE': 'F', 'TRP': 'W', 'TYR': 'Y', 'ASP': 'D', 'ASN': 'N',
    'GLU': 'E', 'LYS': 'K', 'GLN': 'Q', 'MET': 'M', 'SER': 'S',
    'THR': 'T', 'CYS': 'C', 'PRO': 'P', 'HIS': 'H', 'ARG': 'R',
    'UNK': 'X',
}

def infer_atom_type(atom_name):
    """Infer atom type based on atom name for PDB files."""
    atom_name = atom_name.strip().upper()
    if atom_name.startswith('C'):
        return 'C'
    elif atom_name.startswith('O'):
        return 'OA'
    elif atom_name.startswith('N'):
        return 'N'
    elif atom_name.startswith('S'):
        return 'SA'
    elif atom_name.startswith('H'):
        return 'HD'
    else:
        return 'A'  # fallback or generic type

def only_letters(s):
    return re.sub('[^a-zA-Z]', '', s)

def get_chainlist_from_indexfile(chainindex):
    pdb_list = []
    chain_list = []
    with open(chainindex, 'r') as f:
        lines = f.readlines()  
        for line in lines:
            pdb = line.split('\t')[0].strip()
            pdb_list.append(pdb)
            chains = line.split('\t')[1].strip()
            chains = chains.split(';')
            while '' in chains:
                chains.remove('')

            for i, c in enumerate(chains):
                chains[i] = only_letters(c).strip()
            chain_list.append(chains)
            
    return (pdb_list, chain_list)

def find_first_numeric_part(s):
    match = re.search(r'\d+', s)
    return match.group(0) if match else None

def check_res_number(residuelist):
    char_index = []
    for i, res in enumerate(residuelist):
        if res['number'].isnumeric() == False:
            char_index.append(i)
    if len(char_index) == 0:
        return
        
    list_len = len(residuelist) 
    if(char_index[-1] == list_len - 1):
        last_num = find_first_numeric_part(residuelist[-1]['number'])
        residuelist[-1]['number'] = last_num

    for i in reversed(char_index):
        if i == list_len - 1:
            continue

        residuelist[i]['number'] = residuelist[i+1]['number']
        for j in range(i+1, len(residuelist)):
            residuelist[j]['number'] = int(residuelist[j]['number']) + 1

def get_residue_list_from_file(filename, chain_list):
    residue_chain_list = []
    try: 
        with open(filename, 'r') as f:
            lines = f.readlines()

            for chain in chain_list:
                previous_res = -1
                previous_res_type = ''
                atomlist = []
                atom = {}       
                residue = {}
                current_res_chain = ''
                residuelist = []

                for line in lines:
                    atomline = line.split()
                    if atomline[0].strip() != 'ATOM' and atomline[0].strip() != 'TER':
                        continue

                    if line[21] != chain:
                        continue

                    if atomline[0] == 'TER':
                        residue['type'] = aminoacid_abbr.get(previous_res_type, 'X')
                        residue['number'] = previous_res
                        residue['atoms'] = copy.deepcopy(atomlist)
                        residue['chain'] = current_res_chain

                        has_CA = any(atomCA['type'] == 'CA' for atomCA in residue['atoms'])
                        if has_CA:
                            residuelist.append(copy.deepcopy(residue))
                        else:
                            residue.clear()

                        previous_res = -1
                        continue

                    atom['pdbqt_type'] = infer_atom_type(line[12:16].strip())
                    atom['type'] = line[12:16].strip()
                    atom['x'] = line[30:38].strip()
                    atom['y'] = line[38:46].strip()
                    atom['z'] = line[46:54].strip()

                    current_res = line[22:27].strip()
                    current_res_type = line[17:21].strip()
                    current_res_chain = line[21]

                    if current_res != previous_res and previous_res != -1:
                        residue['type'] = aminoacid_abbr.get(previous_res_type, 'X')
                        residue['number'] = previous_res
                        residue['atoms'] = copy.deepcopy(atomlist)
                        residue['chain'] = current_res_chain

                        has_CA = any(atomCA['type'] == 'CA' for atomCA in residue['atoms'])
                        if has_CA:
                            residuelist.append(copy.deepcopy(residue))
                        else:
                            residue.clear()

                        atomlist.clear()
                        atomlist.append(copy.deepcopy(atom))
                    else:
                        atomlist.append(copy.deepcopy(atom))

                    previous_res = current_res
                    previous_res_type = current_res_type

                check_res_number(residuelist)
                residue_chain_list.append(copy.deepcopy(residuelist))

        return residue_chain_list
    except Exception as e:
        print(e)
        return residue_chain_list

# Given 2 parts of protein-protein complex, return the residue list of the complex
# Should be completed 2 residue lists

def get_interaction_residue_pair_new(dis_thred, reslistA, reslistB):
    res_pair = []
    proteinA = copy.deepcopy(reslistA)
    proteinB = copy.deepcopy(reslistB)
    
    for i, resA in enumerate(proteinA):     
        find_match_res_pair = 0
        ca1 = list(filter(lambda x: x['type'] == 'CA', resA['atoms']))[0]
        ca1_x = float(ca1['x'])
        ca1_y = float(ca1['y'])
        ca1_z = float(ca1['z'])
        
        for j, resB in enumerate(proteinB):
            find_match_res_pair = 0
            ca2 = list(filter(lambda x: x['type'] == 'CA', resB['atoms']))[0]
            ca2_x = float(ca2['x'])
            ca2_y = float(ca2['y'])
            ca2_z = float(ca2['z'])  
        
            for atomA in resA['atoms']:
                Ax = float(atomA['x'])
                Ay = float(atomA['y'])
                Az = float(atomA['z'])        

                for atomB in resB['atoms']:
                    Bx = float(atomB['x'])
                    By = float(atomB['y'])
                    Bz = float(atomB['z'])

                    distance = math.sqrt((Ax-Bx)**2 + (Ay-By)**2 + (Az-Bz)**2)
          
                    if distance <= dis_thred:
                        c_alpha_dist = math.sqrt((ca1_x-ca2_x)**2 + (ca1_y-ca2_y)**2 + (ca1_z-ca2_z)**2)
                        res_pair.append((resA, resB, c_alpha_dist))
                        find_match_res_pair = 1
                        break

                if find_match_res_pair == 1:
                    break

    return res_pair

def get_interaction_residue_pair_new_indi(dis_thred, reslistA, reslistB):
    res_pair = []
    proteinA = copy.deepcopy(reslistA)
    proteinB = copy.deepcopy(reslistB)
    
    for i, resA in enumerate(proteinA):     
        find_match_res_pair = 0
        ca1 = list(filter(lambda x: x['type'] == 'CA', resA['atoms']))[0]
        ca1_x = float(ca1['x'])
        ca1_y = float(ca1['y'])
        ca1_z = float(ca1['z'])
        res1_num = resA['number']
        
        for j, resB in enumerate(proteinB):
            find_match_res_pair = 0
            ca2 = list(filter(lambda x: x['type'] == 'CA', resB['atoms']))[0]
            ca2_x = float(ca2['x'])
            ca2_y = float(ca2['y'])
            ca2_z = float(ca2['z'])  
            res2_num = resB['number']
            if res1_num == res2_num:
                continue
        
            for atomA in resA['atoms']:
                Ax = float(atomA['x'])
                Ay = float(atomA['y'])
                Az = float(atomA['z'])        

                for atomB in resB['atoms']:
                    Bx = float(atomB['x'])
                    By = float(atomB['y'])
                    Bz = float(atomB['z'])

                    distance = math.sqrt((Ax-Bx)**2 + (Ay-By)**2 + (Az-Bz)**2)
          
                    if distance <= dis_thred:
     
                        c_alpha_dist = math.sqrt((ca1_x-ca2_x)**2 + (ca1_y-ca2_y)**2 + (ca1_z-ca2_z)**2)
                        res_pair.append((resA, resB, c_alpha_dist))
                        find_match_res_pair = 1
                        break

                if find_match_res_pair == 1:
                    break

    return res_pair

# get the edge index of the graph from the interaction pairs
# only 2 parts in the inter_pairs
# len_pro1 is the length of the first part of the protein-protein complex

def get_edge_index(inter_pairs, len_protein1_fasta):
    source_list = []
    des_list = []

    for pair in inter_pairs:
        source_res = int(pair[0]['number'])
        des_res = int(pair[1]['number']) + len_protein1_fasta
        source_list.append(source_res)
        des_list.append(des_res)

    source = torch.unsqueeze(torch.tensor(source_list), 0)
    des = torch.unsqueeze(torch.tensor(des_list), 0)

    bi_direct_source = torch.cat((source, des), 1)
    bi_direct_edge = torch.cat((des, source), 1)
    edge_index = torch.squeeze(torch.stack((bi_direct_source, bi_direct_edge)))
    
    return edge_index

def get_edge_index_indi(inter_pairs):
    source_list = []
    des_list = []

    for pair in inter_pairs:
        source_res = int(pair[0]['number'])
        des_res = int(pair[1]['number'])
        if source_res != des_res:
            source_list.append(source_res)
            des_list.append(des_res)

    source = torch.unsqueeze(torch.tensor(source_list), 0)
    des = torch.unsqueeze(torch.tensor(des_list), 0)

    edge_index = torch.squeeze(torch.stack((source, des)))
    return edge_index

def adjust_residuelist_num(residuelist):
    index = 0
    for reslist in residuelist:
        for res in reslist:
            res['number'] = index
            index += 1
            
# Helpers
def get_fasta_seq(pdb_tuple):
    fastaA = pdb_tuple[0]
    fastaB = pdb_tuple[1]
    return fastaA, fastaB

def get_distance(x1, y1, z1, x2, y2, z2):
    distance = math.sqrt((x1-x2)**2 + (y1-y2)**2 + (z1-z2)**2)
    return distance

# Models
class AttentiveFPModel(torch.nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels, edge_dim, num_layers, num_timesteps, dropout):
        super(AttentiveFPModel, self).__init__()
        self.model = AttentiveFP(in_channels, hidden_channels, out_channels, edge_dim, num_layers, num_timesteps, dropout)

    def forward(self, data):
        x, edge_index, edge_attr, batch = data.x, data.edge_index, data.edge_attr, data.batch
        return self.model(x, edge_index, edge_attr, batch)

class GraphNetwork(torch.nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels, edge_dim, num_layers, num_timesteps, dropout, linear_out1, linear_out2):
        super(GraphNetwork, self).__init__()
        self.graph1 = AttentiveFPModel(in_channels, hidden_channels, out_channels, edge_dim, num_layers, num_timesteps, dropout)
        self.graph2 = AttentiveFPModel(in_channels, hidden_channels, out_channels, edge_dim, num_layers, num_timesteps, dropout)
        self.graph3 = AttentiveFPModel(in_channels, hidden_channels, out_channels, edge_dim, num_layers, num_timesteps, dropout)
        
        self.fc1 = torch.nn.Linear(out_channels * 3, linear_out1)
        self.fc2 = torch.nn.Linear(linear_out1, linear_out2)

    def forward(self, inter_data, intra_data1, intra_data2):

        inter_graph = self.graph1(inter_data)
        intra_graph1 = self.graph2(intra_data1)
        intra_graph2 = self.graph3(intra_data2)

        x = torch.cat([inter_graph, intra_graph1, intra_graph2], dim=1)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x

def load_esm2_model(use_cache=True):
    """
    Load ESM2 model and tokenizer, with optional caching to avoid reloading.
    
    Args:
        use_cache (bool): Whether to use cached model if available (default: True)
        
    Returns:
        tuple: (tokenizer, model) - ESM2 tokenizer and model
    """
    global _ESM2_MODEL_CACHE, _ESM2_TOKENIZER_CACHE
    
    # Return cached models if available and caching is enabled
    if use_cache and _ESM2_MODEL_CACHE is not None and _ESM2_TOKENIZER_CACHE is not None:
        return _ESM2_TOKENIZER_CACHE, _ESM2_MODEL_CACHE
    
    # Load model and tokenizer
    # Load model and tokenizer with error handling
    try:
        # Set torch to use single thread to avoid segmentation faults
        torch.set_num_threads(1)
        
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            from transformers import AutoTokenizer, EsmModel
            
            # Load tokenizer
            tokenizer = AutoTokenizer.from_pretrained("facebook/esm2_t33_650M_UR50D")
            
            # Load model with explicit CPU device and low memory usage
            model = EsmModel.from_pretrained(
                "facebook/esm2_t33_650M_UR50D",
                torch_dtype=torch.float32
                #low_cpu_mem_usage=False
                #device_map="cpu",
            )
            
            # Ensure model is on CPU?
            # model = model.to('cpu')
        
        model.eval()
        torch.set_grad_enabled(False)
        
        # Cache the models if caching is enabled
        if use_cache:
            _ESM2_MODEL_CACHE = model
            _ESM2_TOKENIZER_CACHE = tokenizer
        
        return tokenizer, model
        
    except Exception as e:
        print(f"Error loading ESM2 model: {str(e)}")
        raise

def clear_esm2_cache():
    """
    Clear the cached ESM2 model and tokenizer to free up memory.
    """
    global _ESM2_MODEL_CACHE, _ESM2_TOKENIZER_CACHE
    _ESM2_MODEL_CACHE = None
    _ESM2_TOKENIZER_CACHE = None
    
    # Force garbage collection
    import gc
    gc.collect()
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        
def run_proaffinity_inference(
        pdbfile, chainindex, weights_path='./model.pkl', temperature=298.15, verbose=False,
        esm_model=None, esm_tokenizer=None
    ):
    """
    Run ProAffinity inference on a given PDB file and chain index.
    
    Args:
        pdbfile (str): Path to the PDB or PDBQT file.
        chainindex (str): List of chains (e.g. "AB,C").
        weights_path (str): Path to the model weights.
        temperature (float): Temperature in Kelvin (default is 298.15) to convert K from.
        verbose (bool): Whether to print verbose output.
        esm_model: Pre-loaded ESM2 model (optional). If None, will load from cache or disk.
        esm_tokenizer: Pre-loaded ESM2 tokenizer (optional). If None, will load from cache or disk.
        
    Returns:
        float: Predicted dG (kJ/mol) value.
    """

    pdbfile = pdbfile
    chainindex = chainindex.split(',')

    # get 2 residue lists from file of one pdb
    # each residue list may contain several res lists, each res list represents a chain

    prot_pair = chainindex

    if len(prot_pair) != 2:
        print('no 2 pro-pro pair!!')

    # Get residues from the PDB file
    if verbose: print(prot_pair)
    residuelistA = get_residue_list_from_file(pdbfile, prot_pair[0])
    residuelistB = get_residue_list_from_file(pdbfile, prot_pair[1])

    adjust_residuelist_num(residuelistA)
    adjust_residuelist_num(residuelistB)

    # Create a sequence for each chain
    if len(residuelistA) == 0 or len(residuelistB) == 0:
        print('no residue list')

    seqA = []

    for reslist in residuelistA: # for each chain in the residuelistA:
        chain_res = []
        for res in reslist:
            chain_res.append(res['type'])
        chain_res = ''.join(chain_res)
        seqA.append(chain_res)

    seqB = []

    for reslist in residuelistB: # for each chain in the residuelistA:
        chain_res = []
        for res in reslist:
            chain_res.append(res['type'])
        chain_res = ''.join(chain_res)
        seqB.append(chain_res)
        
    seqA_len = sum([len(seq) for seq in seqA])

    adjust_residuelist_num(residuelistA)
    adjust_residuelist_num(residuelistB)

    concat_reslistA = list(chain.from_iterable(residuelistA))
    concat_reslistB = list(chain.from_iterable(residuelistB))

    # Generate graphs
    thred = 6
    inter_pairs = get_interaction_residue_pair_new(thred, concat_reslistA, concat_reslistB)
    edge_index = get_edge_index(inter_pairs, seqA_len)
    info_save = (inter_pairs, edge_index, (seqA, seqB))

    # get 2 residue lists from file of one pdb
    # each residue list may contain several res lists, each res list represents a chain

    residuelistA = get_residue_list_from_file(pdbfile, prot_pair[0])
    residuelistB = get_residue_list_from_file(pdbfile, prot_pair[1])

    adjust_residuelist_num(residuelistA)
    adjust_residuelist_num(residuelistB)

    if len(residuelistA) == 0 or len(residuelistB) == 0:
        print('no residue list')

    seqA = []

    for reslist in residuelistA: # for each chain in the residuelistA:
        chain_res = []
        for res in reslist:
            chain_res.append(res['type'])
        chain_res = ''.join(chain_res)
        seqA.append(chain_res)

    seqB = []

    for reslist in residuelistB: # for each chain in the residuelistA:
        chain_res = []
        for res in reslist:
            chain_res.append(res['type'])
        chain_res = ''.join(chain_res)
        seqB.append(chain_res)

    adjust_residuelist_num(residuelistA)
    adjust_residuelist_num(residuelistB)

    concat_reslistA = list(chain.from_iterable(residuelistA))
    concat_reslistB = list(chain.from_iterable(residuelistB))


    thred = 3.5
    intra_pairsA = get_interaction_residue_pair_new_indi(thred, concat_reslistA, concat_reslistA)
    intra_pairsB = get_interaction_residue_pair_new_indi(thred, concat_reslistB, concat_reslistB)

    edge_index1 = get_edge_index_indi(intra_pairsA)
    edge_index2 = get_edge_index_indi(intra_pairsB)

    info_save1 = (intra_pairsA, edge_index1, seqA)
    info_save2 = (intra_pairsB, edge_index2, seqB)
                    
    # Get ESM embedding from the sequences
    # Use pre-loaded model if provided, otherwise load from cache
    if esm_model is None or esm_tokenizer is None:
        tokenizer, model = load_esm2_model(use_cache=True)
    else:
        tokenizer = esm_tokenizer
        model = esm_model

    from torch_geometric.data import Data
    

    atom_type = ['A', 'C', 'OA', 'N', 'NA' 'SA', 'HD']

    atom_pair = ['A_A', 'A_C', 'A_OA', 'A_N', 'A_NA', 'A_SA', 'A_HD', 
                'C_C', 'C_OA', 'C_N', 'C_NA', 'C_SA', 'C_HD',
                'OA_OA', 'OA_N', 'OA_NA', 'OA_SA', 'OA_HD',
                'N_N', 'N_NA', 'N_SA', 'N_HD', 
                'NA_NA', 'NA_SA', 'NA_HD',
                'SA_SA', 'SA_HD',
                'HD_HD']

    bin_number = 10
    type_number = len(atom_pair)
    inter_distance = 6
    intra_distance = 3.5

    seqA, seqB = get_fasta_seq(info_save[2])
    output1_list = []
    output2_list = []

    for fasta in seqA:
        input1 = tokenizer(fasta, return_tensors="pt")
        output1 = model(**input1)
        last_hidden_state1 = output1.last_hidden_state
        last_hidden_state1 = torch.squeeze(last_hidden_state1)
        # get the token from the 2nd to the 2nd last one
        last_hidden_state1 = last_hidden_state1[1:-1]
        output1_list.append(last_hidden_state1)


    for fasta in seqB:
        input2 = tokenizer(fasta, return_tensors="pt")
        output2 = model(**input2)
        last_hidden_state2 = output2.last_hidden_state
        last_hidden_state2 = torch.squeeze(last_hidden_state2)
        # get the token from the 2nd to the 2nd last one
        last_hidden_state2 = last_hidden_state2[1:-1]
        output2_list.append(last_hidden_state2)


    x1 = torch.cat(output1_list, 0)
    x2 = torch.cat(output2_list, 0)
    x = torch.cat((x1, x2), 0)

    pairs = info_save[0]
    edge_index = info_save[1]

    try:
        edge_feature = []
        for pair in pairs:

            edge_encoding = np.zeros(type_number * bin_number)
            residueA = pair[0]
            residueB = pair[1]

            for atom1 in residueA['atoms']:
                x1 = float(atom1['x'])
                y1 = float(atom1['y'])
                z1 = float(atom1['z'])
                type1 = atom1['pdbqt_type']

                for atom2 in residueB['atoms']:
                    x2 = float(atom2['x'])
                    y2 = float(atom2['y'])
                    z2 = float(atom2['z'])
                    type2 = atom2['pdbqt_type']

                    dis = get_distance(x1, y1, z1, x2, y2, z2)

                    bin_n = math.ceil(dis / (inter_distance / bin_number))
                    #print(bin_n)
                    if bin_n > 10:
                        bin_n = 10

                    if type1 + '_' + type2 in atom_pair:
                        pair_type = type1 + '_' + type2       
                    elif type2 + '_' + type1 in atom_pair:
                        pair_type = type2 + '_' + type1
                    else:
                        print('no match type!' + 'type1:' + type1 + 'type2:' + type2)
                        pair_type = 'others'
                        
                    pair_type_index = atom_pair.index(pair_type)
                    encoding_index = (bin_n - 1) * type_number + pair_type_index
                    edge_encoding[encoding_index] = edge_encoding[encoding_index] + 1

            edge_feature.append(torch.from_numpy(edge_encoding))                

        edge_feature = torch.stack(edge_feature, 0) 
        edge_feature = torch.cat((edge_feature, edge_feature), 0)

    except Exception as e:
        print(e)

    data = Data(x=x, edge_index=edge_index, edge_attr=edge_feature)

    # for individual graph

    seq1 = info_save1[2]

    output1_list = []

    for fasta in seq1:
        input1 = tokenizer(fasta, return_tensors="pt")
        output1 = model(**input1)
        last_hidden_state1 = output1.last_hidden_state
        last_hidden_state1 = torch.squeeze(last_hidden_state1)
        # get the token from the 2nd to the 2nd last one
        last_hidden_state1 = last_hidden_state1[1:-1]
        output1_list.append(last_hidden_state1)

    x_indi_1= torch.cat(output1_list, 0)

    pairs = info_save1[0]
    edge_index = info_save1[1]

    try:
        edge_feature = []
        for pair in pairs:

            edge_encoding = np.zeros(type_number * bin_number)
            residueA = pair[0]
            residueB = pair[1]

            for atom1 in residueA['atoms']:
                x1 = float(atom1['x'])
                y1 = float(atom1['y'])
                z1 = float(atom1['z'])
                type1 = atom1['pdbqt_type']

                for atom2 in residueB['atoms']:
                    x2 = float(atom2['x'])
                    y2 = float(atom2['y'])
                    z2 = float(atom2['z'])
                    type2 = atom2['pdbqt_type']

                    dis = get_distance(x1, y1, z1, x2, y2, z2)

                    bin_n = math.ceil(dis / (intra_distance / bin_number))
                    #print(bin_n)
                    if bin_n > 10:
                        bin_n = 10

                    if type1 + '_' + type2 in atom_pair:
                        pair_type = type1 + '_' + type2       
                    elif type2 + '_' + type1 in atom_pair:
                        pair_type = type2 + '_' + type1
                    else:
                        print('no match type!' + 'type1:' + type1 + 'type2:' + type2)
                        pair_type = 'others'
                        
                    pair_type_index = atom_pair.index(pair_type)
                    encoding_index = (bin_n - 1) * type_number + pair_type_index
                    edge_encoding[encoding_index] = edge_encoding[encoding_index] + 1

            edge_feature.append(torch.from_numpy(edge_encoding))                

        edge_feature = torch.stack(edge_feature, 0) 

    except Exception as e:
        print(e)

    data1 = Data(x=x_indi_1, edge_index=edge_index, edge_attr=edge_feature)

    seq2 = info_save2[2]

    output2_list = []

    for fasta in seq2:
        input2 = tokenizer(fasta, return_tensors="pt")
        output1 = model(**input2)
        last_hidden_state2 = output2.last_hidden_state
        last_hidden_state2 = torch.squeeze(last_hidden_state2)
        # get the token from the 2nd to the 2nd last one
        last_hidden_state2 = last_hidden_state2[1:-1]
        output2_list.append(last_hidden_state2)

    x_indi_2= torch.cat(output2_list, 0)

    pairs = info_save2[0]
    edge_index = info_save2[1]

    try:
        edge_feature = []
        for pair in pairs:

            edge_encoding = np.zeros(type_number * bin_number)
            residueA = pair[0]
            residueB = pair[1]

            for atom1 in residueA['atoms']:
                x1 = float(atom1['x'])
                y1 = float(atom1['y'])
                z1 = float(atom1['z'])
                type1 = atom1['pdbqt_type']

                for atom2 in residueB['atoms']:
                    x2 = float(atom2['x'])
                    y2 = float(atom2['y'])
                    z2 = float(atom2['z'])
                    type2 = atom2['pdbqt_type']

                    dis = get_distance(x1, y1, z1, x2, y2, z2)

                    bin_n = math.ceil(dis / (intra_distance / bin_number))
                    #print(bin_n)
                    if bin_n > 10:
                        bin_n = 10

                    if type1 + '_' + type2 in atom_pair:
                        pair_type = type1 + '_' + type2       
                    elif type2 + '_' + type1 in atom_pair:
                        pair_type = type2 + '_' + type1
                    else:
                        print('no match type!' + 'type1:' + type1 + 'type2:' + type2)
                        pair_type = 'others'
                        
                    pair_type_index = atom_pair.index(pair_type)
                    encoding_index = (bin_n - 1) * type_number + pair_type_index
                    edge_encoding[encoding_index] = edge_encoding[encoding_index] + 1

            edge_feature.append(torch.from_numpy(edge_encoding))                

        edge_feature = torch.stack(edge_feature, 0) 

    except Exception as e:
        print(e)


    data2 = Data(x=x_indi_2, edge_index=edge_index, edge_attr=edge_feature)

    graph = data
    graph1 = data1
    graph2 = data2
    graph.edge_attr = graph.edge_attr.float()

    graph1.edge_attr = [attr.float() for attr in graph1.edge_attr]
    graph2.edge_attr = graph2.edge_attr.float()

    datalist_inter = []
    datalist_intra1 = []
    datalist_intra2 = []
    datalist_inter.append(graph)
    datalist_intra1.append(graph1)
    datalist_intra2.append(graph2)

    test_loader_inter = DataLoader(datalist_inter, batch_size=1)
    test_loader_intra1 = DataLoader(datalist_intra1, batch_size=1)
    test_loader_intra2 = DataLoader(datalist_intra2, batch_size=1)

    in_channels = data.num_node_features
    hidden_channels = 256
    out_channels = 64
    linear_out1 = 32   
    linear_out2 = 1
    edge_dim = data.num_edge_features
    num_layers = 3 
    num_timesteps = 2
    dropout = 0.5

    devicename = "cuda" if torch.cuda.is_available() else "cpu"
    device = torch.device(devicename)
    model = GraphNetwork(in_channels, hidden_channels, out_channels, edge_dim, num_layers, num_timesteps, dropout, linear_out1, linear_out2).to(device)

    state_dict = torch.load(weights_path, map_location=devicename, weights_only=True)

    new_state_dict = {}
    for key, value in state_dict.items():
        new_key = key.replace("lin_src", "lin").replace("lin_dst", "lin")
        new_state_dict[new_key] = value

    model.load_state_dict(new_state_dict, strict=False)


    # Assuming model is your GNN model and dataloader is your test dataloader
    model.eval()  # Set the model to evaluation mode
    

    all_predictions = []

    with torch.no_grad():
        # Disable gradient computation during testing
        for batch_inter, batch_intra1, batch_intra2 in zip(test_loader_inter, test_loader_intra1, test_loader_intra2):
            # Assuming batch contains input data 'x' and true values 'y_true'
            batch_inter = batch_inter.to(device)
            batch_intra1 = batch_intra1.to(device)
            batch_intra2 = batch_intra2.to(device)
            if isinstance(batch_intra1.edge_attr, list):
                batch_intra1.edge_attr = torch.stack(batch_intra1.edge_attr, dim=0).to(device) 
            if isinstance(batch_intra2.edge_attr, list):
                batch_intra2.edge_attr = torch.stack(batch_intra2.edge_attr, dim=0).to(device)        # y_true = batch_inter.y
            
            # Get model predictions for the current batch
            y_pred = model(batch_inter, batch_intra1, batch_intra2)
            y_pred = torch.squeeze(y_pred)
            
            # Store predictions and true values
            all_predictions.append(y_pred.cpu().numpy())

    for i in range(len(all_predictions)):
        # Check if the current item is a scalar by examining its dimensionality
        if all_predictions[i].ndim == 0:
            # Convert scalar to a 1D array and update the item in the list
            all_predictions[i] = np.array([all_predictions[i]])

    # Concatenate all predictions and true values
    all_predictions = np.concatenate(all_predictions, axis=0)
    K = 10**(all_predictions[0])
    R = 8.314 / 1000 # kJ/(mol*K)
    dG = -R * temperature * np.log(K)
    return dG

# Example usage
if __name__ == "__main__": 
    run_proaffinity_inference("./pdbfiles/1i4d.pdbqt", "D,AB", verbose=True)





"""
test_reactions_generation_pipeline.py

Unit test for benchmarking the reaction and transformation enumeration pipeline
in the `ode_gen` package using the "8y7s" example molecular graph.

This test verifies the correctness, performance structure, and minimal
functionality of the graph-based species and reaction generation pipeline,
which includes:

1. Subgraph Enumeration:
    - Uses `get_unique_fully_connected_subgraphs` to extract all unique
      fully connected induced subgraphs (species) from the input molecular graph.

2. Dimer Reaction Generation:
    - Applies `find_all_dimer_reactions` to identify all valid pairwise
      species interactions that result in a product graph.
    - Uses `get_broken_edges` to annotate which bonds are removed in each reaction.

3. Transformable Subgraph Detection:
    - Uses `find_all_transformable_subgraph_pairs` to find species pairs
      that differ only by one or more edges, indicating possible
      edge-transformation reactions (e.g., conformational transitions).

4. Structure Validation:
    - Asserts the returned dataframes and metadata dictionary are well-formed.
    - Ensures the number of species, reactions, and transformations is non-zero.
    - Verifies presence of timing information and critical keys in the results.

Test Strategy:
--------------
- Implemented using Python's `unittest` framework.
- Can be run standalone or through `pytest` without modification.
- Designed for reproducibility and integration testing of the core
  reaction graph enumeration pipeline.

Usage:
------
To run as a standalone test:
    python -m unittest tests.benchmark.test_pipeline

To run with pytest:
    pytest tests/benchmark/test_pipeline.py -v

Requires:
---------
- NetworkX-based graph model (`get_example`)
- `ode_gen` modules:
    - `complexes.subcomplexes`
    - `reactions.dimer`
    - `reactions.transformation`
- pandas, platform, and standard Python libraries

Author: yying7@jh.edu
-------
Auto-generated and maintained as part of the ioNERDSS framework
for molecular complex modeling and reaction network generation.
"""

import unittest
from pathlib import Path
import pandas as pd
import json

from ionerdss.model.graph_based.complexes.examples import get_example
from ionerdss.model.graph_based.complexes.subcomplexes import get_unique_fully_connected_subgraphs
from ionerdss.model.graph_based.reactions.dimer import find_all_dimer_reactions, get_broken_edges
from ionerdss.model.graph_based.reactions.transformation import find_all_transformable_subgraph_pairs

from datetime import datetime
import time
import platform

def capture_environment_info():
    return {
        "timestamp": datetime.now().isoformat(),
        "platform": platform.system(),
        "platform_version": platform.version(),
        "python_version": platform.python_version(),
        "hostname": platform.node(),
        "cpu": platform.processor(),
    }

def benchmark_pipeline(graph_name="8y7s"):
    G = get_example(graph_name)
    results = {}
    env_info = capture_environment_info()

    t0 = time.time()
    species = get_unique_fully_connected_subgraphs(G)
    t1 = time.time()

    reactions = find_all_dimer_reactions(species, use_multiprocessing=True)
    t2 = time.time()

    reaction_df = pd.DataFrame([
        {
            "product": list(r[2].nodes),
            "part1": list(r[0]),
            "part2": list(r[1]),
            "bonds_broken": get_broken_edges(r[2], r[0], r[1])
        }
        for r in reactions
    ])

    transformations = find_all_transformable_subgraph_pairs(G, species)
    t3 = time.time()

    transformations_df = pd.DataFrame([
        {
            "monomer_1_nodes": list(t1.nodes),
            "monomer_2_nodes": list(t2.nodes),
            "diff": list(set(t1.edges(data="type")) ^ set(t2.edges(data="type")))
        }
        for t1, t2 in transformations
    ])

    results["timing"] = {
        "subgraph_enumeration": round(t1 - t0, 4),
        "dimer_reactions": round(t2 - t1, 4),
        "transformable_pairs": round(t3 - t2, 4),
        "total": round(t3 - t0, 4),
    }
    results["n_species"] = len(species)
    results["n_reactions"] = len(reaction_df)
    results["n_transformations"] = len(transformations_df)
    results["environment"] = env_info

    return reaction_df, transformations_df, results


class TestBenchmarkPipeline(unittest.TestCase):

    def test_benchmark_output_8y7s(self):
        graph_name = "8y7s"
        reaction_df, transformation_df, metadata = benchmark_pipeline(graph_name)

        # Basic structure checks
        self.assertIsInstance(reaction_df, pd.DataFrame)
        self.assertIsInstance(transformation_df, pd.DataFrame)
        self.assertIsInstance(metadata, dict)

        # Output sanity
        self.assertGreaterEqual(len(reaction_df), 1, "No dimer reactions found")
        self.assertGreaterEqual(len(transformation_df), 0, " Transformable subgraphs found while none expected")
        self.assertIn("timing", metadata)
        self.assertIn("n_species", metadata)

        # Timing keys present
        for key in ["subgraph_enumeration", "dimer_reactions", "transformable_pairs", "total"]:
            self.assertIn(key, metadata["timing"])
            self.assertGreaterEqual(metadata["timing"][key], 0.0)


if __name__ == "__main__":
    unittest.main()



import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
import numpy as np


def gui():
    class Mol:
        def __init__(self, name, sites):
            self.name = name
            self.sites = sites

    class Site:
        def __init__(self, name, states=[]):
            self.name = name
            self.states = states

    class State:
        def __init__(self, name):
            self.name = name

    # Helper function to create labeled entry fields
    def create_labeled_entry_with_hint(
        parent, label_text, hint_text, row, column, width=50
    ):
        label = tk.Label(parent, text=label_text)
        entry = tk.Entry(parent, width=width)
        label.grid(row=row, column=column, sticky="e")
        entry.grid(row=row, column=column + 1)

        entry.insert(0, hint_text)
        entry.bind(
            "<FocusIn>",
            lambda event, entry=entry, hint_text=hint_text: on_focus_in(
                event, entry, hint_text
            ),
        )
        entry.bind(
            "<FocusOut>",
            lambda event, entry=entry, hint_text=hint_text: on_focus_out(
                event, entry, hint_text
            ),
        )

        return entry

    def on_focus_in(event, entry, hint_text):
        if entry.get() == hint_text:
            entry.delete(0, tk.END)

    def on_focus_out(event, entry, hint_text):
        if not entry.get():
            entry.insert(0, hint_text)

    def create_labeled_combobox_entry(
        parent, label_text, options_list, row, column, width=30
    ):
        label = tk.Label(parent, text=label_text)
        combobox = ttk.Combobox(parent, values=options_list, width=width)
        label.grid(row=row, column=column, sticky="e")
        combobox.grid(row=row, column=column + 1)

        return combobox

    def create_labeled_combobox_entry_with_callback(
        parent, label_text, options, row, column, callback, width=30
    ):
        label = tk.Label(parent, text=label_text)
        combobox = ttk.Combobox(parent, values=options, width=width)
        combobox.bind("<<ComboboxSelected>>", callback)
        label.grid(row=row, column=column, sticky="e")
        combobox.grid(row=row, column=column + 1)
        return combobox

    def generate_inp_file():
        with open("parms.inp", "w") as f:
            f.write("start parameters\n")
            f.write(f"    nItr = {nItr_entry.get()}\n")
            f.write(f"    timeStep = {timeStep_entry.get()}\n")
            f.write(f"    timeWrite = {timeWrite_entry.get()}\n")
            f.write(f"    trajWrite = {trajWrite_entry.get()}\n")
            f.write(f"    pdbWrite = {pdbWrite_entry.get()}\n")
            f.write(f"    transitionWrite = {transitionWrite_entry.get()}\n")
            f.write(f"    restartWrite = {restartWrite_entry.get()}\n")
            f.write(f"    checkPoint = {checkPoint_entry.get()}\n")
            f.write(f"    scaleMaxDisplace = {scaleMaxDisplace_entry.get()}\n")
            f.write(f"    overlapSepLimit = {overlapSepLimit_entry.get()}\n")
            f.write(f"    clusterOverlapCheck = {clusterOverlapCheck_entry.get()}\n")
            f.write("end parameters\n\n")

            f.write("start boundaries\n")
            isSphere_value = isSphere_entry.get().strip().lower() == "true"
            if not isSphere_value:
                f.write(f"    WaterBox = {waterBox_entry.get()}\n")
                f.write("    xBCtype = reflect\n")
                f.write("    yBCtype = reflect\n")
                f.write("    zBCtype = reflect\n")
            else:
                f.write("    isSphere = true\n")
                f.write(f"    sphereR  = {sphereR_entry.get()}\n")
            f.write("end boundaries\n\n")

            f.write("start molecules\n")
            for mol_data in added_molecules:
                f.write(f"    {mol_data['name']} : {mol_data['copy_number']}\n")
            f.write("end molecules\n\n")

            f.write("start reactions\n")
            for react_data in reactions:
                f.write(f"{react_data}\n")
            f.write("end reactions\n")
        messagebox.showinfo(
            "File Generated", "The .inp file has been successfully generated."
        )

    def create_mol_file_from_entries():
        mol_name = mol_name_entry.get()
        check_overlap = checkOverlap_entry.get().lower() == "true"
        count_transition = countTransition_entry.get().lower() == "true"
        transition_matrix_size = int(transitionMatrixSize_entry.get())
        is_implicit_lipid = isImplicitLipid_entry.get().lower() == "true"
        if is_implicit_lipid and not added_molecules:
            messagebox.showerror(
                "Error", "Please add the implicit lipid before other mols."
            )
            return
        is_lipid = isLipid_entry.get().lower() == "true"
        D = D_entry.get()
        Dr = Dr_entry.get()

        input_str = coords_entry.get()
        sites_list = input_str.split(";")
        sites = []
        coords_formatted = ""
        for site_str in sites_list:
            if "," not in site_str:
                site_name, site_coord = site_str.split(":")
                x, y, z = site_coord.split()
                coords_formatted += site_name + "\t" + x + "\t" + y + "\t" + z + "\n"
                sites.append(Site(site_name))
            else:
                coord_str, state_str = site_str.split(",")
                coord_str = coord_str.strip()
                state_str = state_str.strip()
                site_name, site_coord = coord_str.split(":")
                x, y, z = site_coord.split()
                coords_formatted += site_name + "\t" + x + "\t" + y + "\t" + z + "\n"
                states = []
                states_str = state_str.split("~")
                for state in states_str:
                    states.append(State(state))
                sites.append(Site(site_name, states=states))

        bonds = [bond.strip() for bond in bonds_entry.get().split(";")]

        file_path = f"{mol_name}.mol"

        mols.append(Mol(mol_name, sites=sites))
        reactant1_entry["values"] = [mol.name for mol in mols]
        reactant2_entry["values"] = [mol.name for mol in mols]
        product1_entry["values"] = [mol.name for mol in mols]
        product2_entry["values"] = [mol.name for mol in mols]

        with open(file_path, "w") as file:
            file.write(f"Name = {mol_name}\n")
            file.write(f"checkOverlap = {check_overlap}\n")
            file.write(f"countTransition = {count_transition}\n")
            file.write(f"transitionMatrixSize = {transition_matrix_size}\n")
            if is_lipid:
                file.write("isLipid = true\n")
            if is_implicit_lipid:
                file.write("isImplicitLipid = true\n")
            file.write(f"D = {D}\n")
            file.write(f"Dr = {Dr}\n")
            file.write("COM    0.0000    0.0000    0.0000\n")

            file.write(f"{coords_formatted}\n")

            file.write("bonds = {}\n".format(len(bonds)))
            for bond in bonds:
                file.write(f"{bond}\n")

            for one_site in sites:
                if one_site.states:
                    all_states_str = f"state={one_site.name}"
                    for one_state in one_site.states:
                        all_states_str += "~" + one_state.name.upper()
                    file.write(f"{all_states_str}\n")

    def add_reaction():
        react_type = reaction_type_entry.get()
        if react_type == "Zeroth Creation: null -> A":
            react_str = "    NULL -> "
            prod_name = product1_entry.get()
            prod_site = product1_site.get()
            prod_state = product1_state.get()
            r = rate_entry.get().strip()
            l = label_entry.get().strip()
            if prod_state == "":
                react_str += prod_name + "(" + prod_site + ")\n"
            else:
                react_str += prod_name + "(" + prod_site + "~" + prod_state + ")\n"
            react_str += "    rate = " + r + "\n"
            if l != "":
                react_str += "    rxnLagbel = " + l + "\n"
            reactions.append(react_str)

        elif react_type == "Destruction: A -> null":
            react_str = "    "
            react_name = reactant1_entry.get()
            react_site = reactant1_site.get()
            react_state = reactant1_state.get()
            r = rate_entry.get().strip()
            l = label_entry.get().strip()
            if react_state == "":
                react_str += react_name + "(" + react_site + ") -> NULL\n"
            else:
                react_str += (
                    react_name + "(" + react_site + "~" + react_state + ") -> NULL\n"
                )
            react_str += "    rate = " + r + "\n"
            if l != "":
                react_str += "    rxnLagbel = " + l + "\n"
            reactions.append(react_str)

        elif react_type == "Statechange: B(b~U) -> B(b~P)":
            react_str = "    "
            react_name = reactant1_entry.get()
            react_site = reactant1_site.get()
            react_state = reactant1_state.get()
            prod_name = product1_entry.get()
            prod_site = product1_site.get()
            prod_state = product1_state.get()
            r = rate_entry.get().strip()
            l = label_entry.get().strip()
            if react_state == "":
                react_str += react_name + "(" + react_site + ") -> "
            else:
                react_str += react_name + "(" + react_site + "~" + react_state + ") -> "
            if prod_state == "":
                react_str += prod_name + "(" + prod_site + ")\n"
            else:
                react_str += prod_name + "(" + prod_site + "~" + prod_state + ")\n"
            react_str += "    rate = " + r + "\n"
            if l != "":
                react_str += "    rxnLagbel = " + l + "\n"
            reactions.append(react_str)

        elif react_type == "Bimolecular Association: (A + B <-> A.B)":
            react1_name = reactant1_entry.get()
            react1_site = reactant1_site.get()
            react1_state = reactant1_state.get()
            react1_req = reactant1_bound_site_requirment.get()
            prod1_name = product1_entry.get()
            prod1_site = product1_site.get()
            prod1_state = product1_state.get()
            react2_name = reactant2_entry.get()
            react2_site = reactant2_site.get()
            react2_state = reactant2_state.get()
            react2_req = reactant2_bound_site_requirment.get()
            prod2_name = product2_entry.get()
            prod2_site = product2_site.get()
            prod2_state = product2_state.get()
            micro_on_rate_str = onRate3Dka_entry.get().strip()
            micro_off_rate_str = offRatekb_entry.get().strip()
            macro_on_rate_str = onRate3DMacro_entry.get().strip()
            macro_off_rate_str = offRateMacro_entry.get().strip()
            kcat_str = kcat_entry.get().strip()
            bindRadSameCom_str = bindRadSameCom_entry.get().strip()
            loopCoopFactor_str = loopCoopFactor_entry.get().strip()
            length3dto2d_str = length3dto2d_entry.get().strip()
            exclude_str = exclude_entry.get().strip()
            coupledLabel_str = coupledLabel_entry.get().strip()
            l = label_entry.get().strip()
            c1_str = coord_com_reactant1_entry.get().strip()
            c2_str = coord_com_reactant2_entry.get().strip()
            p1_str = coord_site_reactant1_entry.get().strip()
            p2_str = coord_site_reactant2_entry.get().strip()

            theta1 = "nan"
            theta2 = "nan"
            phi1 = "nan"
            phi2 = "nan"
            omega = "nan"
            # get c1
            c1_list = c1_str[1:-1].split(",")
            c1 = np.array([float(one) for one in c1_list])
            # get c2
            c2_list = c2_str[1:-1].split(",")
            c2 = np.array([float(one) for one in c2_list])
            # get p1
            p1_list = p1_str[1:-1].split(",")
            p1 = np.array([float(one) for one in p1_list])
            # get p2
            p2_list = p2_str[1:-1].split(",")
            p2 = np.array([float(one) for one in p2_list])
            # get sigma1
            sigma1 = p1 - p2
            # get sigma2
            sigma2 = -sigma1
            # get sigma
            sigma = np.linalg.norm(sigma1)

            v1 = p1 - c1
            v2 = p2 - c2

            # get n1
            n1 = np.cross(v1, sigma1)
            if not np.isclose(np.linalg.norm(n1), 0):
                n1 = n1 / np.linalg.norm(n1)
            else:
                n1 = np.array([0, 0, 1])

            # get n2
            n2 = np.cross(v2, sigma2)
            if not np.isclose(np.linalg.norm(n2), 0):
                n2 = n2 / np.linalg.norm(n2)
            else:
                n2 = np.array([0, 0, 1])

            # get theta1, theta2
            if not np.isclose(np.linalg.norm(v1), 0):
                theta1 = np.arccos(
                    np.dot(v1, sigma1) / (np.linalg.norm(v1) * np.linalg.norm(sigma1))
                )
            else:
                theta1 = "nan"
            if not np.isclose(np.linalg.norm(v2), 0):
                theta2 = np.arccos(
                    np.dot(v2, sigma2) / (np.linalg.norm(v2) * np.linalg.norm(sigma2))
                )
            else:
                theta2 = "nan"

            # get phi1, phi2
            if not np.isclose(np.linalg.norm(np.cross(v1, sigma1)), 0):
                t1 = np.cross(v1, sigma1)
                t2 = np.cross(v1, n1)
                norm_t1 = t1 / np.linalg.norm(t1)
                norm_t2 = t2 / np.linalg.norm(t2)
                phi1 = np.arccos(np.dot(norm_t1, norm_t2))
                if np.dot(np.cross(v1, t1), t2) > 0:
                    phi1 = -phi1
            else:
                phi1 = "nan"

            if not np.isclose(np.linalg.norm(np.cross(v2, sigma2)), 0):
                t1 = np.cross(v2, sigma2)
                t2 = np.cross(v2, n2)
                norm_t1 = t1 / np.linalg.norm(t1)
                norm_t2 = t2 / np.linalg.norm(t2)
                phi2 = np.arccos(np.dot(norm_t1, norm_t2))
                if np.dot(np.cross(v2, t1), t2) > 0:
                    phi2 = -phi2
            else:
                phi1 = "nan"

            # get omega
            if not np.isclose(np.linalg.norm(v1), 0) and not np.isclose(
                np.linalg.norm(v2), 0
            ):
                if not np.isclose(
                    np.linalg.norm(np.cross(v1, sigma1)), 0
                ) and not np.isclose(np.linalg.norm(np.cross(v2, sigma2)), 0):
                    t1 = np.cross(sigma1, v1)
                    t2 = np.cross(sigma1, v2)
                else:
                    t1 = np.cross(sigma1, n1)
                    t2 = np.cross(sigma1, n2)

                omega = np.arccos(
                    np.dot(t1, t2) / (np.linalg.norm(t1) * np.linalg.norm(t2))
                )
                if np.dot(np.cross(sigma1, t1), t2) > 0:
                    omega = -omega
            else:
                omega = "nan"

            react_str = "    "
            if react1_state == "":
                if react1_req == "":
                    react1_str = react1_name + "(" + react1_site + ")"
                    prod1_str = react1_name + "(" + react1_site + "!1)"
                else:
                    react1_str = f"{react1_name}({react1_site}, {react1_req}!*)"
                    prod1_str = f"{react1_name}({react1_site}!1, {react1_req}!*)"
            else:
                if react1_req == "":
                    react1_str = (
                        react1_name + "(" + react1_site + "~" + react1_state + ")"
                    )
                    prod1_str = (
                        react1_name + "(" + react1_site + "~" + react1_state + "!1)"
                    )
                else:
                    react1_str = (
                        f"{react1_name}({react1_site}~{react1_state}, {react1_req}!*)"
                    )
                    prod1_str = (
                        f"{react1_name}({react1_site}~{react1_state}!1, {react1_req}!*)"
                    )
            if react2_state == "":
                if react2_req == "":
                    react2_str = react2_name + "(" + react2_site + ")"
                    prod2_str = react2_name + "(" + react2_site + "!1)"
                else:
                    react2_str = f"{react2_name}({react2_site}, {react2_req}!*)"
                    prod2_str = f"{react2_name}({react2_site}!1, {react2_req}!*)"
            else:
                if react2_req == "":
                    react2_str = (
                        react2_name + "(" + react2_site + "~" + react2_state + ")"
                    )
                    prod2_str = (
                        react2_name + "(" + react2_site + "~" + react2_state + "!1)"
                    )
                else:
                    react2_str = (
                        f"{react2_name}({react2_site}~{react2_state}, {react2_req}!*)"
                    )
                    prod2_str = (
                        f"{react2_name}({react2_site}~{react2_state}!1, {react2_req}!*)"
                    )

            prod_str = prod1_str + "." + prod2_str

            react_str += f"{react1_str} + {react2_str} <-> {prod_str}\n"

            if micro_on_rate_str != "":
                if micro_off_rate_str == "":
                    messagebox.showerror("Error", "Please enter micro off rate.")
                    return
                else:
                    react_str += f"    onRate3Dka = {micro_on_rate_str}\n"
                    react_str += f"    offRatekb = {micro_off_rate_str}\n"
            elif macro_on_rate_str != "":
                if macro_off_rate_str == "":
                    messagebox.showerror("Error", "Please enter macro off rate.")
                    return
                else:
                    react_str += f"    onRate3DMacro = {macro_on_rate_str}\n"
                    react_str += f"    offRateMacro = {macro_off_rate_str}\n"
            else:
                messagebox.showerror(
                    "Error", "Please enter reaction rates (either micro or macro)."
                )
                return

            if kcat_str != "":
                react_str += f"    kcat = {kcat_str}\n"

            react_str += f"    sigma = {sigma}\n"

            if bindRadSameCom_str != "":
                react_str += f"    bindRadSameCom = {bindRadSameCom_str}\n"

            if loopCoopFactor_str != "":
                react_str += f"    loopCoopFactor = {loopCoopFactor_str}\n"

            if length3dto2d_str != "":
                react_str += f"    length3Dto2D = {length3dto2d_str}\n"

            react_str += f"    norm1 = {str(list(n1))}\n"
            react_str += f"    norm2 = {str(list(n2))}\n"
            react_str += (
                "    assocAngles = ["
                + str(theta1)
                + ","
                + str(theta2)
                + ","
                + str(phi1)
                + ","
                + str(phi2)
                + ","
                + str(omega)
                + "]\n"
            )

            if l != "":
                react_str += "    rxnLagbel = " + l + "\n"

            if coupledLabel_str != "":
                react_str += "    coupledRxnLabel = " + coupledLabel_str + "\n"

            if exclude_str != "":
                exclude_value = exclude_str.lower() == "true"
                react_str += f"    excludeVolumeBound = {exclude_value}\n"

            reactions.append(react_str)

        elif react_type == "Michaelis-Menten: A + B <-> A.B -> A + C":
            react1_name = reactant1_entry.get()
            react1_site = reactant1_site.get()
            react1_state = reactant1_state.get()
            react1_req = reactant1_bound_site_requirment.get()
            prod1_name = product1_entry.get()
            prod1_site = product1_site.get()
            prod1_state = product1_state.get()
            react2_name = reactant2_entry.get()
            react2_site = reactant2_site.get()
            react2_state = reactant2_state.get()
            react2_req = reactant2_bound_site_requirment.get()
            prod2_name = product2_entry.get()
            prod2_site = product2_site.get()
            prod2_state = product2_state.get()
            micro_on_rate_str = onRate3Dka_entry.get().strip()
            micro_off_rate_str = offRatekb_entry.get().strip()
            macro_on_rate_str = onRate3DMacro_entry.get().strip()
            macro_off_rate_str = offRateMacro_entry.get().strip()
            kcat_str = kcat_entry.get().strip()
            bindRadSameCom_str = bindRadSameCom_entry.get().strip()
            loopCoopFactor_str = loopCoopFactor_entry.get().strip()
            length3dto2d_str = length3dto2d_entry.get().strip()
            exclude_str = exclude_entry.get().strip()
            coupledLabel_str = coupledLabel_entry.get().strip()
            l = label_entry.get().strip()
            c1_str = coord_com_reactant1_entry.get().strip()
            c2_str = coord_com_reactant2_entry.get().strip()
            p1_str = coord_site_reactant1_entry.get().strip()
            p2_str = coord_site_reactant2_entry.get().strip()

            theta1 = "nan"
            theta2 = "nan"
            phi1 = "nan"
            phi2 = "nan"
            omega = "nan"
            # get c1
            c1_list = c1_str[1:-1].split(",")
            c1 = np.array([float(one) for one in c1_list])
            # get c2
            c2_list = c2_str[1:-1].split(",")
            c2 = np.array([float(one) for one in c2_list])
            # get p1
            p1_list = p1_str[1:-1].split(",")
            p1 = np.array([float(one) for one in p1_list])
            # get p2
            p2_list = p2_str[1:-1].split(",")
            p2 = np.array([float(one) for one in p2_list])
            # get sigma1
            sigma1 = p1 - p2
            # get sigma2
            sigma2 = -sigma1
            # get sigma
            sigma = np.linalg.norm(sigma1)

            v1 = p1 - c1
            v2 = p2 - c2

            # get n1
            n1 = np.cross(v1, sigma1)
            if not np.isclose(np.linalg.norm(n1), 0):
                n1 = n1 / np.linalg.norm(n1)
            else:
                n1 = np.array([0, 0, 1])

            # get n2
            n2 = np.cross(v2, sigma2)
            if not np.isclose(np.linalg.norm(n2), 0):
                n2 = n2 / np.linalg.norm(n2)
            else:
                n2 = np.array([0, 0, 1])

            # get theta1, theta2
            if not np.isclose(np.linalg.norm(v1), 0):
                theta1 = np.arccos(
                    np.dot(v1, sigma1) / (np.linalg.norm(v1) * np.linalg.norm(sigma1))
                )
            else:
                theta1 = "nan"
            if not np.isclose(np.linalg.norm(v2), 0):
                theta2 = np.arccos(
                    np.dot(v2, sigma2) / (np.linalg.norm(v2) * np.linalg.norm(sigma2))
                )
            else:
                theta2 = "nan"

            # get phi1, phi2
            if not np.isclose(np.linalg.norm(np.cross(v1, sigma1)), 0):
                t1 = np.cross(v1, sigma1)
                t2 = np.cross(v1, n1)
                norm_t1 = t1 / np.linalg.norm(t1)
                norm_t2 = t2 / np.linalg.norm(t2)
                phi1 = np.arccos(np.dot(norm_t1, norm_t2))
                if np.dot(np.cross(v1, t1), t2) > 0:
                    phi1 = -phi1
            else:
                phi1 = "nan"

            if not np.isclose(np.linalg.norm(np.cross(v2, sigma2)), 0):
                t1 = np.cross(v2, sigma2)
                t2 = np.cross(v2, n2)
                norm_t1 = t1 / np.linalg.norm(t1)
                norm_t2 = t2 / np.linalg.norm(t2)
                phi2 = np.arccos(np.dot(norm_t1, norm_t2))
                if np.dot(np.cross(v2, t1), t2) > 0:
                    phi2 = -phi2
            else:
                phi1 = "nan"

            # get omega
            if not np.isclose(np.linalg.norm(v1), 0) and not np.isclose(
                np.linalg.norm(v2), 0
            ):
                if not np.isclose(
                    np.linalg.norm(np.cross(v1, sigma1)), 0
                ) and not np.isclose(np.linalg.norm(np.cross(v2, sigma2)), 0):
                    t1 = np.cross(sigma1, v1)
                    t2 = np.cross(sigma1, v2)
                else:
                    t1 = np.cross(sigma1, n1)
                    t2 = np.cross(sigma1, n2)

                omega = np.arccos(
                    np.dot(t1, t2) / (np.linalg.norm(t1) * np.linalg.norm(t2))
                )
                if np.dot(np.cross(sigma1, t1), t2) > 0:
                    omega = -omega
            else:
                omega = "nan"

            react_str = "    "
            if react1_state == "":
                if react1_req == "":
                    react1_str = react1_name + "(" + react1_site + ")"
                    prod1_str = react1_name + "(" + react1_site + "!1)"
                else:
                    react1_str = f"{react1_name}({react1_site}, {react1_req}!*)"
                    prod1_str = f"{react1_name}({react1_site}!1, {react1_req}!*)"
            else:
                if react1_req == "":
                    react1_str = (
                        react1_name + "(" + react1_site + "~" + react1_state + ")"
                    )
                    prod1_str = (
                        react1_name + "(" + react1_site + "~" + react1_state + "!1)"
                    )
                else:
                    react1_str = (
                        f"{react1_name}({react1_site}~{react1_state}, {react1_req}!*)"
                    )
                    prod1_str = (
                        f"{react1_name}({react1_site}~{react1_state}!1, {react1_req}!*)"
                    )
            if react2_state == "":
                if react2_req == "":
                    react2_str = react2_name + "(" + react2_site + ")"
                    prod2_str = react2_name + "(" + react2_site + "!1)"
                else:
                    react2_str = f"{react2_name}({react2_site}, {react2_req}!*)"
                    prod2_str = f"{react2_name}({react2_site}!1, {react2_req}!*)"
            else:
                if react2_req == "":
                    react2_str = (
                        react2_name + "(" + react2_site + "~" + react2_state + ")"
                    )
                    prod2_str = (
                        react2_name + "(" + react2_site + "~" + react2_state + "!1)"
                    )
                else:
                    react2_str = (
                        f"{react2_name}({react2_site}~{react2_state}, {react2_req}!*)"
                    )
                    prod2_str = (
                        f"{react2_name}({react2_site}~{react2_state}!1, {react2_req}!*)"
                    )

            prod_str = prod1_str + "." + prod2_str

            react_str += f"{react1_str} + {react2_str} <-> {prod_str}\n"

            if micro_on_rate_str != "":
                if micro_off_rate_str == "":
                    messagebox.showerror("Error", "Please enter micro off rate.")
                    return
                else:
                    react_str += f"    onRate3Dka = {micro_on_rate_str}\n"
                    react_str += f"    offRatekb = {micro_off_rate_str}\n"
            elif macro_on_rate_str != "":
                if macro_off_rate_str == "":
                    messagebox.showerror("Error", "Please enter macro off rate.")
                    return
                else:
                    react_str += f"    onRate3DMacro = {macro_on_rate_str}\n"
                    react_str += f"    offRateMacro = {macro_off_rate_str}\n"
            else:
                messagebox.showerror(
                    "Error", "Please enter reaction rates (either micro or macro)."
                )
                return

            if kcat_str != "":
                react_str += f"    kcat = {kcat_str}\n"
            else:
                messagebox.showerror(
                    "Error", "Please specify kcat for Michaelis-Menten."
                )
                return

            react_str += f"    sigma = {sigma}\n"

            if bindRadSameCom_str != "":
                react_str += f"    bindRadSameCom = {bindRadSameCom_str}\n"

            if loopCoopFactor_str != "":
                react_str += f"    loopCoopFactor = {loopCoopFactor_str}\n"

            if length3dto2d_str != "":
                react_str += f"    length3Dto2D = {length3dto2d_str}\n"

            react_str += f"    norm1 = {str(list(n1))}\n"
            react_str += f"    norm2 = {str(list(n2))}\n"
            react_str += (
                "    assocAngles = ["
                + str(theta1)
                + ","
                + str(theta2)
                + ","
                + str(phi1)
                + ","
                + str(phi2)
                + ","
                + str(omega)
                + "]\n"
            )

            if l != "":
                react_str += "    rxnLagbel = " + l + "\n"

            if coupledLabel_str != "":
                react_str += "    coupledRxnLabel = " + coupledLabel_str + "\n"
            else:
                messagebox.showerror(
                    "Error", "Please specify coupled reaction for Michaelis-Menten."
                )
                return

            if exclude_str != "":
                exclude_value = exclude_str.lower() == "true"
                react_str += f"    excludeVolumeBound = {exclude_value}\n"

            reactions.append(react_str)

        elif react_type == "Bimolecular Statechange: A + B(b~U) -> A + B(b~P)":
            react1_name = reactant1_entry.get()
            react1_site = reactant1_site.get()
            react1_state = reactant1_state.get()
            react1_req = reactant1_bound_site_requirment.get()
            prod1_name = product1_entry.get()
            prod1_site = product1_site.get()
            prod1_state = product1_state.get()
            prod1_req = product1_bound_site_requirment.get()
            react2_name = reactant2_entry.get()
            react2_site = reactant2_site.get()
            react2_state = reactant2_state.get()
            react2_req = reactant2_bound_site_requirment.get()
            prod2_name = product2_entry.get()
            prod2_site = product2_site.get()
            prod2_state = product2_state.get()
            prod2_req = product2_bound_site_requirment.get()
            micro_on_rate_str = onRate3Dka_entry.get().strip()
            micro_off_rate_str = offRatekb_entry.get().strip()
            macro_on_rate_str = onRate3DMacro_entry.get().strip()
            macro_off_rate_str = offRateMacro_entry.get().strip()
            kcat_str = kcat_entry.get().strip()
            bindRadSameCom_str = bindRadSameCom_entry.get().strip()
            loopCoopFactor_str = loopCoopFactor_entry.get().strip()
            length3dto2d_str = length3dto2d_entry.get().strip()
            exclude_str = exclude_entry.get().strip()
            coupledLabel_str = coupledLabel_entry.get().strip()
            l = label_entry.get().strip()
            c1_str = coord_com_reactant1_entry.get().strip()
            c2_str = coord_com_reactant2_entry.get().strip()
            p1_str = coord_site_reactant1_entry.get().strip()
            p2_str = coord_site_reactant2_entry.get().strip()

            theta1 = "nan"
            theta2 = "nan"
            phi1 = "nan"
            phi2 = "nan"
            omega = "nan"
            # get c1
            c1_list = c1_str[1:-1].split(",")
            c1 = np.array([float(one) for one in c1_list])
            # get c2
            c2_list = c2_str[1:-1].split(",")
            c2 = np.array([float(one) for one in c2_list])
            # get p1
            p1_list = p1_str[1:-1].split(",")
            p1 = np.array([float(one) for one in p1_list])
            # get p2
            p2_list = p2_str[1:-1].split(",")
            p2 = np.array([float(one) for one in p2_list])
            # get sigma1
            sigma1 = p1 - p2
            # get sigma2
            sigma2 = -sigma1
            # get sigma
            sigma = np.linalg.norm(sigma1)

            v1 = p1 - c1
            v2 = p2 - c2

            # get n1
            n1 = np.cross(v1, sigma1)
            if not np.isclose(np.linalg.norm(n1), 0):
                n1 = n1 / np.linalg.norm(n1)
            else:
                n1 = np.array([0, 0, 1])

            # get n2
            n2 = np.cross(v2, sigma2)
            if not np.isclose(np.linalg.norm(n2), 0):
                n2 = n2 / np.linalg.norm(n2)
            else:
                n2 = np.array([0, 0, 1])

            # get theta1, theta2
            if not np.isclose(np.linalg.norm(v1), 0):
                theta1 = np.arccos(
                    np.dot(v1, sigma1) / (np.linalg.norm(v1) * np.linalg.norm(sigma1))
                )
            else:
                theta1 = "nan"
            if not np.isclose(np.linalg.norm(v2), 0):
                theta2 = np.arccos(
                    np.dot(v2, sigma2) / (np.linalg.norm(v2) * np.linalg.norm(sigma2))
                )
            else:
                theta2 = "nan"

            # get phi1, phi2
            if not np.isclose(np.linalg.norm(np.cross(v1, sigma1)), 0):
                t1 = np.cross(v1, sigma1)
                t2 = np.cross(v1, n1)
                norm_t1 = t1 / np.linalg.norm(t1)
                norm_t2 = t2 / np.linalg.norm(t2)
                phi1 = np.arccos(np.dot(norm_t1, norm_t2))
                if np.dot(np.cross(v1, t1), t2) > 0:
                    phi1 = -phi1
            else:
                phi1 = "nan"

            if not np.isclose(np.linalg.norm(np.cross(v2, sigma2)), 0):
                t1 = np.cross(v2, sigma2)
                t2 = np.cross(v2, n2)
                norm_t1 = t1 / np.linalg.norm(t1)
                norm_t2 = t2 / np.linalg.norm(t2)
                phi2 = np.arccos(np.dot(norm_t1, norm_t2))
                if np.dot(np.cross(v2, t1), t2) > 0:
                    phi2 = -phi2
            else:
                phi1 = "nan"

            # get omega
            if not np.isclose(np.linalg.norm(v1), 0) and not np.isclose(
                np.linalg.norm(v2), 0
            ):
                if not np.isclose(
                    np.linalg.norm(np.cross(v1, sigma1)), 0
                ) and not np.isclose(np.linalg.norm(np.cross(v2, sigma2)), 0):
                    t1 = np.cross(sigma1, v1)
                    t2 = np.cross(sigma1, v2)
                else:
                    t1 = np.cross(sigma1, n1)
                    t2 = np.cross(sigma1, n2)

                omega = np.arccos(
                    np.dot(t1, t2) / (np.linalg.norm(t1) * np.linalg.norm(t2))
                )
                if np.dot(np.cross(sigma1, t1), t2) > 0:
                    omega = -omega
            else:
                omega = "nan"

            react_str = "    "
            if react1_state == "":
                if react1_req == "":
                    react1_str = react1_name + "(" + react1_site + ")"
                else:
                    react1_str = f"{react1_name}({react1_site}, {react1_req}!*)"
            else:
                if react1_req == "":
                    react1_str = (
                        react1_name + "(" + react1_site + "~" + react1_state + ")"
                    )
                else:
                    react1_str = (
                        f"{react1_name}({react1_site}~{react1_state}, {react1_req}!*)"
                    )
            if react2_state == "":
                if react2_req == "":
                    react2_str = react2_name + "(" + react2_site + ")"
                else:
                    react2_str = f"{react2_name}({react2_site}, {react2_req}!*)"
            else:
                if react2_req == "":
                    react2_str = (
                        react2_name + "(" + react2_site + "~" + react2_state + ")"
                    )
                else:
                    react2_str = (
                        f"{react2_name}({react2_site}~{react2_state}, {react2_req}!*)"
                    )

            if prod1_state == "":
                if prod1_req == "":
                    prod1_str = prod1_name + "(" + prod1_site + ")"
                else:
                    prod1_str = f"{prod1_name}({prod1_site}, {prod1_req}!*)"
            else:
                if prod1_req == "":
                    prod1_str = prod1_name + "(" + prod1_site + "~" + prod1_state + ")"
                else:
                    prod1_str = (
                        f"{prod1_name}({prod1_site}~{prod1_state}, {prod1_req}!*)"
                    )
            if prod2_state == "":
                if prod2_req == "":
                    prod2_str = prod2_name + "(" + prod2_site + ")"
                else:
                    prod2_str = f"{prod2_name}({prod2_site}, {prod2_req}!*)"
            else:
                if prod2_req == "":
                    prod2_str = prod2_name + "(" + prod2_site + "~" + prod2_state + ")"
                else:
                    prod2_str = (
                        f"{prod2_name}({prod2_site}~{prod2_state}, {prod2_req}!*)"
                    )

            react_str += f"{react1_str} + {react2_str} <-> {prod1_str} + {prod2_str}\n"

            if micro_on_rate_str != "":
                if micro_off_rate_str == "":
                    messagebox.showerror("Error", "Please enter micro off rate.")
                    return
                else:
                    react_str += f"    onRate3Dka = {micro_on_rate_str}\n"
                    react_str += f"    offRatekb = {micro_off_rate_str}\n"
            elif macro_on_rate_str != "":
                if macro_off_rate_str == "":
                    messagebox.showerror("Error", "Please enter macro off rate.")
                    return
                else:
                    react_str += f"    onRate3DMacro = {macro_on_rate_str}\n"
                    react_str += f"    offRateMacro = {macro_off_rate_str}\n"
            else:
                messagebox.showerror(
                    "Error", "Please enter reaction rates (either micro or macro)."
                )
                return

            if kcat_str != "":
                react_str += f"    kcat = {kcat_str}\n"

            react_str += f"    sigma = {sigma}\n"

            if bindRadSameCom_str != "":
                react_str += f"    bindRadSameCom = {bindRadSameCom_str}\n"

            if loopCoopFactor_str != "":
                react_str += f"    loopCoopFactor = {loopCoopFactor_str}\n"

            if length3dto2d_str != "":
                react_str += f"    length3Dto2D = {length3dto2d_str}\n"

            react_str += f"    norm1 = {str(list(n1))}\n"
            react_str += f"    norm2 = {str(list(n2))}\n"
            react_str += (
                "    assocAngles = ["
                + str(theta1)
                + ","
                + str(theta2)
                + ","
                + str(phi1)
                + ","
                + str(phi2)
                + ","
                + str(omega)
                + "]\n"
            )

            if l != "":
                react_str += "    rxnLagbel = " + l + "\n"

            if coupledLabel_str != "":
                react_str += "    coupledRxnLabel = " + coupledLabel_str + "\n"

            if exclude_str != "":
                exclude_value = exclude_str.lower() == "true"
                react_str += f"    excludeVolumeBound = {exclude_value}\n"

            reactions.append(react_str)

        elif react_type == "Unimolecular Creation: A -> A + B":
            react1_name = reactant1_entry.get()
            react1_site = reactant1_site.get()
            react1_state = reactant1_state.get()
            react1_req = reactant1_bound_site_requirment.get()
            prod1_name = product1_entry.get()
            prod1_site = product1_site.get()
            prod1_state = product1_state.get()
            prod1_req = product1_bound_site_requirment.get()
            react2_name = reactant2_entry.get()
            react2_site = reactant2_site.get()
            react2_state = reactant2_state.get()
            react2_req = reactant2_bound_site_requirment.get()
            prod2_name = product2_entry.get()
            prod2_site = product2_site.get()
            prod2_state = product2_state.get()
            prod2_req = product2_bound_site_requirment.get()
            micro_on_rate_str = onRate3Dka_entry.get().strip()
            micro_off_rate_str = offRatekb_entry.get().strip()
            macro_on_rate_str = onRate3DMacro_entry.get().strip()
            macro_off_rate_str = offRateMacro_entry.get().strip()
            kcat_str = kcat_entry.get().strip()
            bindRadSameCom_str = bindRadSameCom_entry.get().strip()
            loopCoopFactor_str = loopCoopFactor_entry.get().strip()
            length3dto2d_str = length3dto2d_entry.get().strip()
            exclude_str = exclude_entry.get().strip()
            coupledLabel_str = coupledLabel_entry.get().strip()
            l = label_entry.get().strip()
            r = rate_entry.get().strip()

            react_str = "    "
            if react1_state == "":
                if react1_req == "":
                    react1_str = react1_name + "(" + react1_site + ")"
                else:
                    react1_str = f"{react1_name}({react1_site}, {react1_req}!*)"
            else:
                if react1_req == "":
                    react1_str = (
                        react1_name + "(" + react1_site + "~" + react1_state + ")"
                    )
                else:
                    react1_str = (
                        f"{react1_name}({react1_site}~{react1_state}, {react1_req}!*)"
                    )

            if prod1_state == "":
                if prod1_req == "":
                    prod1_str = prod1_name + "(" + prod1_site + ")"
                else:
                    prod1_str = f"{prod1_name}({prod1_site}, {prod1_req}!*)"
            else:
                if prod1_req == "":
                    prod1_str = prod1_name + "(" + prod1_site + "~" + prod1_state + ")"
                else:
                    prod1_str = (
                        f"{prod1_name}({prod1_site}~{prod1_state}, {prod1_req}!*)"
                    )
            if prod2_state == "":
                if prod2_req == "":
                    prod2_str = prod2_name + "(" + prod2_site + ")"
                else:
                    prod2_str = f"{prod2_name}({prod2_site}, {prod2_req}!*)"
            else:
                if prod2_req == "":
                    prod2_str = prod2_name + "(" + prod2_site + "~" + prod2_state + ")"
                else:
                    prod2_str = (
                        f"{prod2_name}({prod2_site}~{prod2_state}, {prod2_req}!*)"
                    )

            react_str += f"{react1_str} <-> {prod1_str} + {prod2_str}\n"

            if r != "":
                react_str += f"    rate = {r}\n"

            if l != "":
                react_str += "    rxnLagbel = " + l + "\n"

            if coupledLabel_str != "":
                react_str += "    coupledRxnLabel = " + coupledLabel_str + "\n"

            reactions.append(react_str)

        messagebox.showinfo("Reaction Added", "Reaction Added.")
        return

    def add_molecule():
        mol_name = mol_name_entry.get().strip()
        copy_num = mol_copy_num_entry.get().strip()
        concentration = mol_concentration_entry.get().strip()

        if not mol_name:
            messagebox.showerror("Error", "Please enter a molecule name.")
            return

        if not copy_num and not concentration:
            messagebox.showerror(
                "Error", "Please enter either copy number or concentration."
            )
            return

        if copy_num and concentration:
            messagebox.showerror(
                "Error",
                f"Please enter either copy number {copy_num} OR concentration {concentration}, not both.",
            )
            return

        if concentration:
            # Convert concentration to copy number
            isSphere_value = isSphere_entry.get().strip().lower() == "true"
            volume = 0
            if not isSphere_value:
                dimensions_str = waterBox_entry.get().strip()
                dimensions = [int(dim) for dim in dimensions_str.strip("[]").split(",")]
                x, y, z = dimensions
                volume = x * y * z / 1e9  # um3
            else:
                r = int(sphereR_entry.get().strip())
                volume = 4 / 3 * 3.1415927 * r**3 / 1e9  # um3
            copy_num = int(
                float(concentration) / 1000000 * volume / 1000000000000000 * 6.02e23
            )

        percentage = states_entry.get().strip()
        if not percentage:
            mol_data = {"name": mol_name, "copy_number": copy_num}
            added_molecules.append(mol_data)
            molecules_listbox.insert(tk.END, f"{mol_name} : {copy_num}")
        else:
            state_str_list = percentage.split(",")
            new_copy_number_str_list = []
            for one_state in state_str_list:
                num, name = one_state.split()
                num = int(float(num) * int(copy_num))
                new_copy_number_str_list.append(str(num) + " " + name)
                new_copy_number_str = ",".join(new_copy_number_str_list)
            mol_data = {"name": mol_name, "copy_number": new_copy_number_str}
            added_molecules.append(mol_data)
            molecules_listbox.insert(tk.END, f"{mol_name} : {new_copy_number_str}")

        create_mol_file_from_entries()
        messagebox.showinfo(
            "Molecule Added",
            f"The {mol_name}.mol file has been successfully generated.",
        )

    def update_combobox_options(combobox, options):
        combobox["values"] = options
        combobox.set("")  # Clear the current selection

    def update_reaction_site_options(selected_molecule, sites_combobox):
        sites_options = []
        for mol in mols:
            if mol.name == selected_molecule:
                for site in mol.sites:
                    sites_options.append(site.name)
        if sites_options:
            update_combobox_options(sites_combobox, sites_options)
        else:
            update_combobox_options(sites_combobox, [])

    def update_reaction_state_options(selected_site, states_combobox):
        states_options = []
        for mol in mols:
            for site in mol.sites:
                if site.name == selected_site:
                    for state in site.states:
                        states_options.append(state.name)
        if states_options:
            update_combobox_options(states_combobox, states_options)
        else:
            update_combobox_options(states_combobox, [])

    def on_reactant1_selected(event):
        selected_mol = reactant1_entry.get()
        update_reaction_site_options(selected_mol, reactant1_site)
        update_reaction_site_options(selected_mol, reactant1_bound_site_requirment)

    def on_reactant2_selected(event):
        selected_mol = reactant2_entry.get()
        update_reaction_site_options(selected_mol, reactant2_site)
        update_reaction_site_options(selected_mol, reactant2_bound_site_requirment)

    def on_product1_selected(event):
        selected_mol = product1_entry.get()
        update_reaction_site_options(selected_mol, product1_site)

    def on_product2_selected(event):
        selected_mol = product2_entry.get()
        update_reaction_site_options(selected_mol, product2_site)

    def on_reactant1_site_selected(event):
        selected_site = reactant1_site.get()
        update_reaction_state_options(selected_site, reactant1_state)

    def on_reactant2_site_selected(event):
        selected_site = reactant2_site.get()
        update_reaction_state_options(selected_site, reactant2_state)

    def on_product1_site_selected(event):
        selected_site = product1_site.get()
        update_reaction_state_options(selected_site, product1_state)

    def on_product2_site_selected(event):
        selected_site = product2_site.get()
        update_reaction_state_options(selected_site, product2_state)

    root = tk.Tk()
    root.title("Simulation Parameters")

    # Create a tabbed interface
    tab_parent = ttk.Notebook(root)

    # Parameters tab
    parameters_tab = ttk.Frame(tab_parent)
    tab_parent.add(parameters_tab, text="Parameters")

    nItr_entry = create_labeled_entry_with_hint(
        parameters_tab, "Enter total iteration (steps):", "1000", 0, 0
    )
    timeStep_entry = create_labeled_entry_with_hint(
        parameters_tab, "Enter timeStep (us):", "0.1", 1, 0
    )
    timeWrite_entry = create_labeled_entry_with_hint(
        parameters_tab, "Enter how often to write output (steps):", "10", 2, 0
    )
    trajWrite_entry = create_labeled_entry_with_hint(
        parameters_tab, "Enter how often to write traj (steps):", "100", 3, 0
    )
    pdbWrite_entry = create_labeled_entry_with_hint(
        parameters_tab, "Enter how often to write pdb (steps):", "100", 4, 0
    )
    transitionWrite_entry = create_labeled_entry_with_hint(
        parameters_tab,
        "Enter how often to write transition matrix (steps):",
        "100",
        5,
        0,
    )
    restartWrite_entry = create_labeled_entry_with_hint(
        parameters_tab, "Enter how often to write restart file (steps):", "100", 6, 0
    )
    checkPoint_entry = create_labeled_entry_with_hint(
        parameters_tab, "Enter how often to write check points (steps):", "100", 7, 0
    )
    scaleMaxDisplace_entry = create_labeled_entry_with_hint(
        parameters_tab, "Enter threshold to reject association (<RMSD>):", "100", 8, 0
    )
    overlapSepLimit_entry = create_labeled_entry_with_hint(
        parameters_tab,
        "Enter threshold for overlap check between COMs (nm):",
        "0.1",
        9,
        0,
    )
    clusterOverlapCheck_entry = create_labeled_entry_with_hint(
        parameters_tab,
        "Enter whether overlap is checked based on cluster (bool):",
        "False",
        10,
        0,
    )

    # Boundaries tab
    boundaries_tab = ttk.Frame(tab_parent)
    tab_parent.add(boundaries_tab, text="Boundaries")

    isSphere_entry = create_labeled_entry_with_hint(
        boundaries_tab, "Enter if the system is spherical (bool):", "False", 0, 0
    )
    waterBox_entry = create_labeled_entry_with_hint(
        boundaries_tab,
        "Enter the dimensions of the system if it is a box [x,y,z] (nm):",
        "[100,100,100]",
        1,
        0,
    )
    sphereR_entry = create_labeled_entry_with_hint(
        boundaries_tab,
        "Enter the radius of the sphere if the system is a sphere (nm):",
        "100",
        2,
        0,
    )

    # Molecules tab
    molecules_tab = ttk.Frame(tab_parent)
    tab_parent.add(molecules_tab, text="Molecules")
    # Create widgets for Molecules tab
    mol_name_entry = create_labeled_entry_with_hint(
        molecules_tab, "Enter molecule name:", "", 0, 0
    )
    mol_copy_num_entry = create_labeled_entry_with_hint(
        molecules_tab, "Enter copy number:", "", 1, 0
    )
    mol_concentration_entry = create_labeled_entry_with_hint(
        molecules_tab,
        "Enter concentration if copy number is not provided (uM):",
        "",
        2,
        0,
    )
    checkOverlap_entry = create_labeled_entry_with_hint(
        molecules_tab,
        "Enter boolean value to check overlap for this molecule type:",
        "True",
        3,
        0,
    )
    countTransition_entry = create_labeled_entry_with_hint(
        molecules_tab,
        "Enter boolean value to track transition matrix for this molecule type::",
        "True",
        4,
        0,
    )
    transitionMatrixSize_entry = create_labeled_entry_with_hint(
        molecules_tab,
        "Enter transition matrix size for this molecule type, which should be larger than the size of the formed largest complex:",
        "500",
        5,
        0,
    )
    isImplicitLipid_entry = create_labeled_entry_with_hint(
        molecules_tab, "Enter if it is isImplicitLipid (bool):", "False", 6, 0
    )
    isLipid_entry = create_labeled_entry_with_hint(
        molecules_tab, "Enter if it is lipid (bool):", "False", 7, 0
    )
    D_entry = create_labeled_entry_with_hint(
        molecules_tab,
        "Enter translational diffusion constants ([Dx,Dy,Dz] (um2/s)):",
        "[10,10,10]",
        8,
        0,
    )
    Dr_entry = create_labeled_entry_with_hint(
        molecules_tab,
        "Enter rotational diffusion constants ([Drx,Dry,Drz] (rad2/us)):",
        "[0.1,0,1,0.1]",
        9,
        0,
    )
    coords_entry = create_labeled_entry_with_hint(
        molecules_tab,
        "Enter coords and states of all binding sites, states (one upper letter) are optional (c1: 1 1 1, U~P;c2: 2 2 2; c3: 3 3 3) (nm):",
        "",
        10,
        0,
    )
    bonds_entry = create_labeled_entry_with_hint(
        molecules_tab, "Enter bonds between sites (com c1;com c2;com c3):", "", 11, 0
    )
    states_entry = create_labeled_entry_with_hint(
        molecules_tab,
        "Enter percentage of different initial states (0.8 (sitename~statename1), 0.2 (sitename~statename2)):",
        "",
        12,
        0,
    )
    add_mol_button = tk.Button(molecules_tab, text="Add Molecule", command=add_molecule)
    molecules_listbox = tk.Listbox(molecules_tab)

    add_mol_button.grid(row=13, column=0, columnspan=2)
    molecules_listbox.grid(row=14, column=0, columnspan=2)

    # List to store added molecules
    added_molecules = []

    # all the mols in the system
    mols = []

    # Reactions tab
    reactions_tab = ttk.Frame(tab_parent)
    tab_parent.add(reactions_tab, text="Reactions")

    mol_names = [mol.name for mol in mols]
    reactant1_entry = create_labeled_combobox_entry_with_callback(
        reactions_tab,
        "Select reactant 1 if it is not creation reaction:",
        mol_names,
        0,
        0,
        on_reactant1_selected,
        width=10,
    )
    reactant1_site = create_labeled_combobox_entry_with_callback(
        reactions_tab,
        "Select reactant 1 site:",
        [],
        0,
        2,
        on_reactant1_site_selected,
        width=10,
    )
    reactant1_state = create_labeled_combobox_entry(
        reactions_tab, "Select reactant 1 state (optional):", [], 1, 0, width=10
    )
    reactant1_bound_site_requirment = create_labeled_combobox_entry(
        reactions_tab,
        "Select the site of reactant 1 that needs to be already \nbound for the reaction to occur (optional):",
        [],
        1,
        2,
        width=10,
    )
    reactant2_entry = create_labeled_combobox_entry_with_callback(
        reactions_tab,
        "Select reactant 2 if it is not creation reaction:",
        mol_names,
        2,
        0,
        on_reactant2_selected,
        width=10,
    )
    reactant2_site = create_labeled_combobox_entry_with_callback(
        reactions_tab,
        "Select reactant 2 site:",
        [],
        2,
        2,
        on_reactant2_site_selected,
        width=10,
    )
    reactant2_state = create_labeled_combobox_entry(
        reactions_tab, "Select reactant 2 state (optional):", [], 3, 0, width=10
    )
    reactant2_bound_site_requirment = create_labeled_combobox_entry(
        reactions_tab,
        "Select the site of reactant2 that needs to be already \nbound for the reaction to occur (optional):",
        [],
        3,
        2,
        width=10,
    )
    product1_entry = create_labeled_combobox_entry_with_callback(
        reactions_tab,
        "Select product 1 if it is not bimolecular reaction or destruction reaction:",
        mol_names,
        4,
        0,
        on_product1_selected,
        width=10,
    )
    product1_site = create_labeled_combobox_entry_with_callback(
        reactions_tab,
        "Select product 1 site:",
        [],
        4,
        2,
        on_product1_site_selected,
        width=10,
    )
    product1_bound_site_requirment = create_labeled_combobox_entry(
        reactions_tab,
        "Select the site of product 1 that needs to be already \nbound for the reaction to occur (optional):",
        [],
        5,
        0,
        width=10,
    )
    product1_state = create_labeled_combobox_entry(
        reactions_tab, "Select product 1 state (optional):", [], 5, 2, width=10
    )
    product2_entry = create_labeled_combobox_entry_with_callback(
        reactions_tab,
        "Select product 2 if it is not bimolecular reaction or destruction reaction:",
        mol_names,
        6,
        0,
        on_product2_selected,
        width=10,
    )
    product2_site = create_labeled_combobox_entry_with_callback(
        reactions_tab,
        "Select product 2 site:",
        [],
        6,
        2,
        on_product2_site_selected,
        width=10,
    )
    product2_state = create_labeled_combobox_entry(
        reactions_tab, "Select product 2 state (optional):", [], 7, 0, width=10
    )
    product2_bound_site_requirment = create_labeled_combobox_entry(
        reactions_tab,
        "Select the site of product 2 that needs to be already \nbound for the reaction to occur (optional):",
        [],
        7,
        2,
        width=10,
    )
    coord_com_reactant1_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter reactant 1's com coords \nfor bimolecular reaction ([x,y,z](nm)):",
        "[0,0,0]",
        8,
        0,
        width=10,
    )
    coord_site_reactant1_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter reactant 1's reaction site coords \nfor bimolecular reaction ([x,y,z](nm)):",
        "",
        8,
        2,
        width=10,
    )
    coord_com_reactant2_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter reactant 2's com coords \n(reactant 1 and 2 are in the same coordinate system) ([x,y,z](nm)):",
        "",
        9,
        0,
        width=10,
    )
    coord_site_reactant2_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter reactant 2's reaction site coords: ([x,y,z](nm)):",
        "",
        9,
        2,
        width=10,
    )
    onRate3Dka_entry = create_labeled_entry_with_hint(
        reactions_tab, "Enter microscopic on rate (nm^3 us^-1):", "", 10, 0, width=10
    )
    offRatekb_entry = create_labeled_entry_with_hint(
        reactions_tab, "Enter microscopic off rate (s^-1):", "", 10, 2, width=10
    )
    onRate3DMacro_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter macroscopic on rate if micro not provided (uM^-1 s^-1):",
        "",
        11,
        0,
        width=10,
    )
    offRateMacro_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter macroscopic off rate if micro not provided (s^-1):",
        "",
        11,
        2,
        width=10,
    )
    rate_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter rate for creation / destruction / uni statechange \n/ uni creation (M s^-1 / s^-1 / s^-1 / s^-1):",
        "",
        12,
        0,
        width=10,
    )
    length3dto2d_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter length scale to convert 3D rate to 2D rate\n for bimolecular association (nm):",
        "10",
        13,
        0,
        width=10,
    )
    bindRadSameCom_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter distance betweeen two reactants to force reaction \nwithinin the same complex for bimolecular association (sigma):",
        "1.1",
        13,
        2,
        width=10,
    )
    loopCoopFactor_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter scale factor of rate when closing loops \nfor bimolecular association (ka):",
        "1",
        14,
        0,
        width=10,
    )
    label_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter reaction label, used if you want to couple a \ndifferent reaction to this one:",
        "",
        14,
        2,
        width=10,
    )
    coupledLabel_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter coupled reaction label, used if you allow \nthe completion of this reaction to cause another reaction:",
        "",
        15,
        0,
        width=10,
    )
    kcat_entry = create_labeled_entry_with_hint(
        reactions_tab, "Enter rate of the coupled reaction (s^-1):", "", 15, 2, width=10
    )
    exclude_entry = create_labeled_entry_with_hint(
        reactions_tab,
        "Enter if check exclude volume for bound sites:",
        "False",
        16,
        0,
        width=10,
    )
    reaction_type = [
        "Bimolecular Association: (A + B <-> A.B)",
        "Zeroth Creation: null -> A",
        "Destruction: A -> null",
        "Michaelis-Menten: A + B <-> A.B -> A + C",
        "Unimolecular Creation: A -> A + B",
        "Bimolecular Statechange: A + B(b~U) -> A + B(b~P)",
        "Statechange: B(b~U) -> B(b~P)",
    ]
    reaction_type_entry = create_labeled_combobox_entry(
        reactions_tab, "Select reaction type:", reaction_type, 17, 0
    )
    add_reaction_button = tk.Button(
        reactions_tab, text="Add Reaction", command=add_reaction
    )
    add_reaction_button.grid(row=18, column=1, pady=10)

    # all reactions added
    reactions = []

    tab_parent.pack(expand=True, fill=tk.BOTH)

    generate_button = tk.Button(
        root, text="Generate .inp File", command=generate_inp_file
    )
    generate_button.pack()

    root.mainloop()


def main():
    gui()


if __name__ == "__main__":
    main()



"""
ionerdss.model.components.instances

Dependencies:
    - numpy: For coordinate arrays and mathematical operations
    - ionerdss.model.components.types: Template classes (MoleculeType, InterfaceType)

See Also:
    ionerdss.model.components.types: Template definitions for molecules and interfaces
"""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import Dict, List, Optional

import numpy as np

from ionerdss.model.components.types import MoleculeType, InterfaceType

#------------ helper ---------------
# make sure that mol name does not incude _, which is the separater
# for interface name, e.g. A_12 and B would have interface A-12_B_1
def _replace_underscore_with_dash(input_string : str) -> str:
    """Replace underscores with dashes in molecule names.
    
    Ensures molecule names don't contain underscores which are used as
    separators in interface naming conventions (e.g., "A_B_1").
    
    Args:
        input_string: The string to process.
        
    Returns:
        String with all underscores replaced by dashes.
        
    Example:
        "A_12" becomes "A-12" to avoid conflicts with interface naming.
    """
    return input_string.replace("_", "-")
    
# -------------- dataclasses ------------

@dataclass
class InterfaceInstance:
    """Represents a specific instance of a molecular interface in a simulation.
    
    An InterfaceInstance is a concrete realization of an InterfaceType, representing
    an actual binding site on a specific molecule instance with current spatial
    coordinates and binding state. It maintains references to its partner interface
    and parent molecule instance.
    
    Attributes:
        this_mol_name: Name of the molecule containing this interface (e.g., "A").
        partner_mol_name: Name of the target binding partner molecule (e.g., "B").
        interface_index: Numeric index for distinguishing multiple interfaces
            of the same type on one molecule.
        partner_interface: The InterfaceInstance on the partner molecule that
            this interface is bound to or can bind to.
        interface_type: Reference to the InterfaceType template that defines
            this interface's properties.
        this_mol: Reference to the MoleculeInstance containing this interface.
        absolute_coord: Current absolute Cartesian coordinates of the interface.
        residues: List of residue indices that participate in this interface.
        energy: Current binding energy of this interface. Defaults to -1.0.
        signature: Dictionary containing interface-specific runtime properties.
    """
    # Required fields (no defaults) must come first
    absolute_coord: np.ndarray # absolute cartesian coordinates

    # Optional fields with defaults
    partner_interface: Optional['InterfaceInstance'] = None
    interface_type: Optional[InterfaceType] = None
    this_mol: Optional['MoleculeInstance'] = None
    this_mol_name: str = "unnamed"
    partner_mol_name: str = "unnamed"
    interface_index: int = 0
    residues: List[int] = field(default_factory=list)
    energy: Optional[float] = -1.0
    signature: Dict = field(default_factory=dict)

    def _convert_numpy_types(self, obj):
        """Convert NumPy types to native Python types for JSON serialization."""
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, (np.float32, np.float64)):
            return float(obj)
        elif isinstance(obj, (np.int32, np.int64)):
            return int(obj)
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, dict):
            return {key: self._convert_numpy_types(value) for key, value in obj.items()}
        elif isinstance(obj, (list, tuple)):
            return [self._convert_numpy_types(item) for item in obj]
        else:
            return obj

    def get_name(self) -> str:
        """Return the formatted interface identifier string.
        
        Constructs the interface name using the format:
        "{this_mol_name}_{partner_mol_name}_{interface_index}"
        
        Returns:
            The interface identifier string (e.g., "A_B_1").
        """
        return self.this_mol_name + "_" +\
            self.partner_mol_name + "_" +\
            str(self.interface_index)

    def set_name(self, new_name: str):
        """Parse and set interface identifiers from a formatted name string.
        
        Parses a name string in the format "{mol}_{partner}_{index}" and
        updates the corresponding instance attributes.
        
        Args:
            new_name: Interface name string to parse (e.g., "A_B_1").
            
        Raises:
            IndexError: If the name string doesn't contain exactly 3 
                underscore-separated components.
            ValueError: If the interface index cannot be converted to integer.
        """
        substrings = new_name.split("_")
        self.this_mol_name = substrings[0]
        self.partner_mol_name = substrings[1]
        self.interface_index = int(substrings[2])

    def to_dict(self) -> dict:
        """Convert the InterfaceInstance to a dictionary representation."""
        data = {
            "name": self.get_name(),
            "type": self.interface_type.get_name() if self.interface_type else "unknown",
            "coord": self.absolute_coord,
            "residues": list(self.residues) if hasattr(self, 'residues') else [],
            "energy": self.energy,
        }
        return self._convert_numpy_types(data)

    @classmethod
    def from_dict(cls, d: dict) -> "InterfaceInstance":
        """Create an InterfaceInstance from a dictionary.
        
        Deserializes interface instance data from a dictionary format, typically
        loaded from JSON or other storage formats. Missing optional fields
        are filled with default values.
        
        Args:
            d: Dictionary containing interface data. Required key is
                'absolute_coords'. Optional keys include 'this_mol_name',
                'partner_mol_name', 'interface_index', 'this_mol',
                'partner_interface', 'interface_type', 'energy', and 'signature'.
        
        Returns:
            New InterfaceInstance populated with the dictionary data.
            
        Raises:
            KeyError: If required key 'absolute_coords' is missing.
        """
        return cls(
            absolute_coord=np.array(d["absolute_coords"]),
            partner_interface=d.get("partner_interface", None),
            interface_type=d.get("interface_type", None),
            this_mol=d.get("this_mol", None),
            this_mol_name=d.get("this_mol_name", "unnamed"),
            partner_mol_name=d.get("partner_mol_name", "unnamed"),
            interface_index=d.get("interface_index", 0),
            residues=d.get("residues", []),
            energy=d.get("energy", -1.0),
            signature=d.get("signature", {}),
        )
        
    # Add these methods to InterfaceInstance class
    def __hash__(self) -> int:
        """Make InterfaceInstance hashable for use as dictionary keys."""
        return hash((self.this_mol_name, self.partner_mol_name, self.interface_index))

    def __eq__(self, other) -> bool:
        """Define equality for InterfaceInstance objects."""
        if not isinstance(other, InterfaceInstance):
            return False
        return (self.this_mol_name == other.this_mol_name and 
                self.partner_mol_name == other.partner_mol_name and 
                self.interface_index == other.interface_index)

@dataclass
class MoleculeInstance:
    """Represents a specific instance of a molecule in a simulation.
    
    A MoleculeInstance is a concrete realization of a MoleculeType, representing
    an actual molecule in the simulation with current position, orientation, and
    binding state. It maintains a map of its interface instances and their
    current binding partners.
    
    Note : in implementation keys of interface_neighbors_map are
    changed from InterfaceInstances into string due to the instances
    containing mutable numpy arrays.
    
    Attributes:
        name: Unique identifier for this molecule instance.
        interfaces_neighbors_map: Dictionary mapping interface strings
            to their currently bound neighbor MoleculeInstance objects.
        molecule_type: Reference to the MoleculeType template that defines
            this molecule's properties.
        norm: Orientation vector (relative coordinates) defining molecular
            rotation state.
        com: Center of mass coordinates (absolute position) of the molecule.
    """
    # Required fields first (no defaults)
    name: str
    norm: np.ndarray
    ref1: np.ndarray
    ref2: np.ndarray
    com: np.ndarray # absolute coord

    # Optional fields with defaults
    molecule_type: Optional[MoleculeType] = None
    interfaces_neighbors_map: Dict = field(default_factory=dict)
    
    def get_name(self):
        """Adding a get_name method to conform to the InterfaceInstance class"""
        return self.name

    def _convert_numpy_types(self, obj):
        """Convert NumPy types to native Python types for JSON serialization."""
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, (np.float32, np.float64)):
            return float(obj)
        elif isinstance(obj, (np.int32, np.int64)):
            return int(obj)
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, dict):
            return {key: self._convert_numpy_types(value) for key, value in obj.items()}
        elif isinstance(obj, (list, tuple)):
            return [self._convert_numpy_types(item) for item in obj]
        else:
            return obj

    # In MoleculeInstance class
    def to_dict(self) -> dict:
        """Convert the MoleculeInstance to a dictionary representation."""
        # Convert interfaces_neighbors_map to serializable format
        interfaces_list = []
        for interface_instance, partner_mol in self.interfaces_neighbors_map.items():
            interfaces_list.append({
                "interface_name": interface_instance.get_name(),
                "partner_molecule": partner_mol.name if partner_mol else "unknown"
            })
        
        data = {
            "name": self.name,
            "type": self.molecule_type.name if self.molecule_type else "unknown",
            "com": self.com,
            "norm": self.norm,
            "ref1": self.ref1,
            "ref2": self.ref2,
            "interfaces": interfaces_list
        }
        return self._convert_numpy_types(data)


    # In MoleculeInstance class
    @classmethod
    def from_dict(cls, d: dict) -> "MoleculeInstance":
        """Create a MoleculeInstance from a dictionary.
        
        Note: The interfaces_neighbors_map will be empty and needs to be 
        rebuilt by the system's _rebuild_cross_references() method.
        """
        return cls(
            name=_replace_underscore_with_dash(d.get("name", "unnamed")),
            norm=np.array(d.get("norm", [0.0, 0.0, 1.0])),
            ref1=np.array(d.get("ref1", [1.0, 0.0, 0.0])),
            ref2=np.array(d.get("ref2", [0.0, 0.0, -1.0])),
            com=np.array(d.get("coord", [0.0, 0.0, 0.0])),
            molecule_type=d.get("molecule_type", None),
            interfaces_neighbors_map={}  # Will be rebuilt by system
        )


#!/usr/bin/env python3.10
"""
This script reads the parms.inp file, finds the molecules, 
and generates a new parms.inp file with titration reactions added.
"""

import os
import re
import random
import argparse

def parse_mol_file(mol_file_path):
    """
    Parses a NERDSS .mol file to extract binding interfaces.
    Returns a list of interface names.
    """
    interfaces = []
    try:
        with open(mol_file_path, 'r') as f:
            lines = f.readlines()
            
        in_coords_section = False
        for line in lines:
            line = line.strip()
            if not line or line.startswith('#'):
                # Check if we are entering the coords section
                if "coordinates" in line.lower() or "coords" in line.lower():
                    in_coords_section = True
                continue
                
            # If we hit another section header like # bonds, stop parsing coords
            if line.startswith('#') and in_coords_section:
                in_coords_section = False
                continue
                
            if in_coords_section or len(line.split()) >= 4:
                # Typically, coordinate lines have 4 columns: Name X Y Z
                parts = line.split()
                if len(parts) >= 1:
                    name = parts[0]
                    # Exclude COM and REF
                    if name.lower() not in ['com', 'ref'] and not name.startswith('#'):
                        # Additional safety check to avoid grabbing non-coordinate lines
                        try:
                            if len(parts) >= 4:
                                float(parts[1])
                                float(parts[2])
                                float(parts[3])
                                interfaces.append(name)
                        except ValueError:
                            pass
                            
    except FileNotFoundError:
        print(f"Warning: {mol_file_path} not found.")
    
    return interfaces

def generate_titration_reactions(parms_file, output_file, rate_min=0.00001, rate_max=0.0001):
    """
    Reads parms.inp, extracts molecules, finds their .mol files, 
    generates titration reactions, and writes to parms_titrate.inp.
    """
    base_dir = os.path.dirname(os.path.abspath(parms_file))
    if not base_dir:
        base_dir = "."
        
    molecules = []
    
    # Read parms.inp
    with open(parms_file, 'r') as f:
        parms_lines = f.readlines()
        
    # Extract molecules
    in_molecules_section = False
    for line in parms_lines:
        line_strip = line.strip()
        if line_strip == "start molecules":
            in_molecules_section = True
            continue
        elif line_strip == "end molecules":
            in_molecules_section = False
            break
            
        if in_molecules_section and line_strip and not line_strip.startswith('#'):
            # Format is typically "A0 : 29"
            mol_name = line_strip.split(':')[0].strip()
            molecules.append(mol_name)
            
    print(f"Found molecules: {molecules}")
    
    # Generate titration reactions text
    titration_reactions_text = ["\n    # Titration Reactions\n"]
    
    for mol in molecules:
        mol_file = os.path.join(base_dir, f"{mol}.mol")
        interfaces = parse_mol_file(mol_file)
        
        # Format the reaction
        # 0 -> A0(a0b01, a0c1, a0g1, a0h1, a0i1)
        if interfaces:
            interfaces_str = ", ".join(interfaces)
            reaction_str = f"    0 -> {mol}({interfaces_str})"
        else:
            reaction_str = f"    0 -> {mol}()"
            
        titration_reactions_text.append(reaction_str)
        
        # Generate random rate
        rate = random.uniform(rate_min, rate_max)
        titration_reactions_text.append(f"    onRate3Dka = {rate:.6f} # M/s\n")
        
    # Write new file
    with open(output_file, 'w') as f:
        for line in parms_lines:
            f.write(line)
            
            if line.strip() == "start reactions":
                # Write titration reactions immediately after "start reactions"
                for titr_line in titration_reactions_text:
                    f.write(titr_line + "\n")
                    
    print(f"Successfully wrote {output_file} with {len(molecules)} titration reactions.")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Add titration reactions to NERDSS parms.inp")
    parser.add_argument("input_file", help="Path to input parms.inp")
    parser.add_argument("-o", "--output", default="parms_titrate.inp", help="Output file name (default: parms_titrate.inp)")
    parser.add_argument("--min-rate", type=float, default=0.00001, help="Minimum onRate3Dka (default: 0.00001)")
    parser.add_argument("--max-rate", type=float, default=0.0001, help="Maximum onRate3Dka (default: 0.0001)")
    
    args = parser.parse_args()
    
    # Handle output path to be in same dir as input if only filename is provided
    output_path = args.output
    if not os.path.dirname(output_path):
        output_path = os.path.join(os.path.dirname(os.path.abspath(args.input_file)), args.output)
        
    generate_titration_reactions(args.input_file, output_path, args.min_rate, args.max_rate)



"""Specific implementation of Platonic solids coordinate generation."""

from abc import ABC, abstractmethod
from typing import List, Tuple
import math
import numpy as np
from .geometry import distance, mid_pt

class PlatonicSolidGenerator(ABC):
    """Abstract base class for Platonic solid generators."""

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def num_sites(self) -> int:
        pass

    @abstractmethod
    def _get_vertices(self, radius: float) -> List[List[float]]:
        """Calculate vertices for the solid."""
        pass

    @abstractmethod
    def _get_face_indices(self) -> List[Tuple[int, ...]]:
        """Return list of vertex indices for each face."""
        pass

    @property
    @abstractmethod
    def angle_indices(self) -> Tuple[Tuple[int, int], ...]:
        """Return indices for angle calculation (theta1, theta2, phi1, phi2)."""
        pass

    @abstractmethod
    def _get_reduction_angle(self) -> float:
        """Return the angle used for leg reduction."""
        pass

    def generate_coordinates(self, radius: float, sigma: float) -> List[List[np.ndarray]]:
        """
        Generate coordinates for ALL faces.
        Returns: List of [COM, leg1, leg2, ..., Normal] for each face.
        """
        vertices = self._get_vertices(radius)
        face_indices_list = self._get_face_indices()
        
        # Calculate reduction params
        angle = self._get_reduction_angle()
        denom = 2 * math.sin(angle / 2)
        red_len = sigma / denom
        
        all_faces_coords = []
        
        for face_indices in face_indices_list:
            face_verts = [vertices[i] for i in face_indices]
            
            # 1. Calculate Face COM
            com = np.mean(face_verts, axis=0)
            
            # 2. Calculate Legs (Edge Midpoints)
            legs = []
            num_verts = len(face_verts)
            for i in range(num_verts):
                p1 = face_verts[i]
                p2 = face_verts[(i + 1) % num_verts]
                legs.append(mid_pt(p1, p2))
                
            # 3. Reduce Legs
            reduced_legs = []
            for leg in legs:
                dist = distance(com, leg)
                if dist == 0:
                    ratio = 1 
                else:
                    ratio = 1 - red_len / dist
                
                leg_red = (np.array(leg) - com) * ratio + com
                reduced_legs.append(leg_red)
                
            # 4. Normal (pointing towards origin)
            # Original code used -COM.
            normal = -com
            
            # Assemble list: [COM, leg1, leg2, ..., Normal]
            # Note: Normal is NOT usually part of the flat list used for angle indices,
            # but it IS returned by legacy `input_coord`.
            # Legacy `reduced_coord` returns: [COM, leg1_red, leg2_red...]
            # Legacy `input_coord` adds Normal at the end.
            
            # We will return the structure expected by `angle_cal` (via indices):
            # angle_cal expects points.
            # And `PlatonicSolids.py` expects to extract [COM, legs..., Normal] for the final MoleculeType.
            
            face_data = [com] + reduced_legs + [normal]
            all_faces_coords.append(face_data)
            
        return all_faces_coords

class CubeGenerator(PlatonicSolidGenerator):
    @property
    def name(self): return "cube"
    @property
    def num_sites(self): return 4
    
    def _get_reduction_angle(self):
        return math.acos(0) # 90 degrees

    def _get_vertices(self, radius):
        scaler = radius / (3**0.5)
        return [
            [scaler, scaler, scaler],      # v0
            [-scaler, scaler, scaler],     # v1
            [scaler, -scaler, scaler],     # v2
            [scaler, scaler, -scaler],     # v3
            [-scaler, -scaler, scaler],    # v4
            [scaler, -scaler, -scaler],    # v5
            [-scaler, scaler, -scaler],    # v6
            [-scaler, -scaler, -scaler]    # v7
        ]

    def _get_face_indices(self):
        # 0, 3, 5, 2
        # 0, 3, 6, 1
        # 0, 1, 4, 2
        # 7, 4, 1, 6
        # 7, 4, 2, 5
        # 7, 6, 3, 5
        return [
            (0, 3, 5, 2),
            (0, 3, 6, 1),
            (0, 1, 4, 2),
            (7, 4, 1, 6),
            (7, 4, 2, 5),
            (7, 6, 3, 5)
        ]

    @property
    def angle_indices(self):
        return ((0, 0), (0, 1), (1, 0), (1, 1))

class DodecahedronGenerator(PlatonicSolidGenerator):
    @property
    def name(self): return "dode"
    @property
    def num_sites(self): return 5
    
    def _get_reduction_angle(self):
        m = (1 + 5**0.5) / 2
        return 2 * math.atan(m)

    def _get_vertices(self, radius):
        scaler = radius / (3**0.5)
        m = (1 + 5**0.5) / 2
        
        # Vertices 1-20 mapped to 0-19
        # Code used 1-based names V1..V20
        coords = [
            [0, m, 1/m],    # V1 -> 0
            [0, m, -1/m],   # V2 -> 1
            [0, -m, 1/m],   # V3 -> 2
            [0, -m, -1/m],  # V4 -> 3
            [1/m, 0, m],    # V5 -> 4
            [1/m, 0, -m],   # V6 -> 5
            [-1/m, 0, m],   # V7 -> 6
            [-1/m, 0, -m],  # V8 -> 7
            [m, 1/m, 0],    # V9 -> 8
            [m, -1/m, 0],   # V10 -> 9
            [-m, 1/m, 0],   # V11 -> 10
            [-m, -1/m, 0],  # V12 -> 11
            [1, 1, 1],      # V13 -> 12
            [1, 1, -1],     # V14 -> 13
            [1, -1, 1],     # V15 -> 14
            [1, -1, -1],    # V16 -> 15
            [-1, 1, 1],     # V17 -> 16
            [-1, 1, -1],    # V18 -> 17
            [-1, -1, 1],    # V19 -> 18
            [-1, -1, -1]    # V20 -> 19
        ]
        return [[c * scaler for c in coord] for coord in coords]

    def _get_face_indices(self):
        # Indices adjusted to 0-based from generic inspection
        # 1. 6, 18, 2, 14, 4 -> V7, V19, V3, V15, V5 -> Indices 6, 18, 2, 14, 4
        return [
            (6, 18, 2, 14, 4),
            (6, 4, 12, 0, 16),
            (4, 14, 9, 8, 12),
            (6, 18, 11, 10, 16),
            (14, 2, 3, 15, 9),
            (18, 11, 19, 3, 2),
            (16, 10, 17, 1, 0),
            (12, 0, 1, 13, 8),
            (7, 17, 10, 11, 19),
            (5, 13, 8, 9, 15),
            (3, 19, 7, 5, 15),
            (1, 17, 7, 5, 13)
        ]

    @property
    def angle_indices(self):
        return ((0, 0), (0, 3), (4, 0), (4, 1))

class IcosahedronGenerator(PlatonicSolidGenerator):
    @property
    def name(self): return "icos"
    @property
    def num_sites(self): return 3
    
    def _get_reduction_angle(self):
        return math.acos(-(5**0.5)/3)

    def _get_vertices(self, radius):
        phi = (1 + np.sqrt(5)) / 2
        verts = np.array([
            [0,  1,  phi],
            [0, -1,  phi],
            [0,  1, -phi],
            [0, -1, -phi],
            [1,  phi, 0],
            [-1,  phi, 0],
            [1, -phi, 0],
            [-1, -phi, 0],
            [phi, 0,  1],
            [-phi, 0,  1],
            [phi, 0, -1],
            [-phi, 0, -1],
        ])
        return verts * (radius / np.linalg.norm(verts[0]))

    def _get_face_indices(self):
        return [
            (0, 1, 8),
            (0, 8, 4),
            (0, 4, 5),
            (0, 5, 9),
            (0, 9, 1),
            (1, 9, 7),
            (1, 7, 6),
            (1, 6, 8),
            (8, 6, 10),
            (8, 10, 4),
            (4, 10, 2),
            (4, 2, 5),
            (5, 2, 11),
            (5, 11, 9),
            (9, 11, 7),
            (7, 11, 3),
            (7, 3, 6),
            (6, 3, 10),
            (2, 10, 3),
            (2, 3, 11)
        ]

    @property
    def angle_indices(self):
        return ((0, 0), (0, 3), (1, 0), (1, 1))

class OctahedronGenerator(PlatonicSolidGenerator):
    @property
    def name(self): return "octa"
    @property
    def num_sites(self): return 3
    
    def _get_reduction_angle(self):
        # Angle for Octahedron leg reduction?
        # Assuming standard dihedral (109.47) or similar logic. 
        # Standard implementation used specific logic.
        # Verified earlier: Octa also used leg reduction logic?
        # Wait, I didn't verify Octa angle.
        # Assuming typical: acos(-1/3) = 109.47 deg = tetrahedral angle.
        return math.acos(-1/3) 

    def _get_vertices(self, radius):
        r = radius
        # v0..v5
        coords = [
            [ r,  0,  0],  # 0
            [-r,  0,  0],  # 1
            [ 0,  r,  0],  # 2
            [ 0, -r,  0],  # 3
            [ 0,  0,  r],  # 4
            [ 0,  0, -r],  # 5
        ]
        return coords

    def _get_face_indices(self):
        return [
            (4, 0, 2),
            (4, 2, 1),
            (4, 1, 3),
            (4, 3, 0),
            (5, 2, 0),
            (5, 1, 2),
            (5, 3, 1),
            (5, 0, 3),
        ]

    @property
    def angle_indices(self):
        return ((0, 0), (0, 3), (1, 0), (1, 1))

class TetrahedronGenerator(PlatonicSolidGenerator):
    @property
    def name(self): return "tetr"
    @property
    def num_sites(self): return 3
    
    def _get_reduction_angle(self):
        # Tetrahedron angle. 
        # acos(1/3) is approx 70.5 deg.
        return math.acos(1/3)

    def _get_vertices(self, radius):
        s = radius / np.sqrt(3)
        
        coords = [
            [s,  s,  s],   # v0
            [s, -s, -s],   # v1
            [-s, s, -s],   # v2
            [-s, -s, s]    # v3
        ]
        return coords

    def _get_face_indices(self):
        return [
            (0, 1, 2), (0, 2, 3), (0, 1, 3), (1, 2, 3)
        ]

    @property
    def angle_indices(self):
        return ((0, 0), (0, 3), (1, 0), (1, 1))
